import asyncio
import base64
import datetime
import enum
import hashlib
import hmac
import json
import urllib.parse
from dataclasses import dataclass
from textwrap import shorten
from typing import Any, Dict, List, Optional, Sequence

import aiohttp
import dateutil.parser
from pydantic import Field, FilePath, PositiveFloat, model_validator

from avtdl.core import utils
from avtdl.core.config import Plugins
from avtdl.core.interfaces import MAX_REPR_LEN, Record, RuntimeContext
from avtdl.core.monitors import BaseFeedMonitor, BaseFeedMonitorConfig, BaseFeedMonitorEntity
from avtdl.core.utils import Fmt


@Plugins.register('rplay', Plugins.kind.ASSOCIATED_RECORD)
@Plugins.register('rplay.user', Plugins.kind.ASSOCIATED_RECORD)
class RplayRecord(Record):
    """Ongoing livestream on RPLAY"""
    url: str
    """livestream url"""
    title: str
    """stream title"""
    description: str
    """stream description"""
    thumbnail_url: Optional[str]
    """url of the stream thumbnail"""
    start: datetime.datetime
    """time of the stream start"""
    user_id: str
    """user's oid"""
    creator_id: str
    """creatorOid, unique id used in channel and livestream urls"""
    name: str
    """visible name of the user"""
    avatar_url: str
    """link to the user's avatar"""
    restream_platform: Optional[str] = None
    """for restream, platform the stream is being hosted on"""
    restream_url: Optional[str] = None
    """for restream, url of the stream on the source platform"""
    playlist_url: Optional[str] = None
    """for native stream, link to the underlying hls playlist if was retrieved successfully.
    Might be invalid even when present in case of insufficient permissions or network error"""

    def __str__(self):
        restream = f'\n(restream from {self.restream_url})\n' if self.restream_url else ''
        playlist = f'\n({self.playlist_url})\n' if self.playlist_url else ''
        text = f'[{self.name}] live since {Fmt.date(self.start)}\n{self.title}\n{self.url}{restream}{playlist}'
        return text

    def __repr__(self):
        title = shorten(self.title, MAX_REPR_LEN)
        return f'RplayRecord(creator_id={self.creator_id}, start={self.start}, title={title})'

    def get_uid(self) -> str:
        return f'{self.creator_id}:{int(self.start.timestamp() * 1000)}'

    def discord_embed(self) -> dict:
        return {
            'title': self.title,
            'url': self.url,
            'description': self.restream_url or self.playlist_url or None,
            'image': {'url': self.thumbnail_url},
            'color': None,
            'author': {'name': self.name, 'url': None, 'icon_url': self.avatar_url},
            'timestamp': self.start.isoformat(),
            'footer': {'text': None},
            'fields': []
        }


@Plugins.register('rplay', Plugins.kind.ACTOR_CONFIG)
class RplayMonitorConfig(BaseFeedMonitorConfig):
    pass


@Plugins.register('rplay', Plugins.kind.ACTOR_ENTITY)
class RplayMonitorEntity(BaseFeedMonitorEntity):
    creators: Optional[List[str]] = None
    """list of IDs of the users to monitor. Entity with no IDs will report on every user going live"""
    update_interval: PositiveFloat = 300
    """how often the monitored channel will be checked, in seconds"""
    selected_for_update: bool = Field(exclude=True, default=False)
    """internal variable, used to mark one entity that actually makes network request to update queues"""
    url: str = Field(exclude=True, default=None)
    """url is not used, all streams are checked through the same api endpoing"""
    cookies_file: Optional[FilePath] = Field(exclude=True, default=None)
    """cookies are not used to log in"""
    adjust_update_interval: bool = Field(exclude=True, default=True)
    """rplay api endpoints doesn't use caching headers"""


@Plugins.register('rplay', Plugins.kind.ACTOR)
class RplayMonitor(BaseFeedMonitor):
    """
    Monitor livestreams on RPLAY

    Monitors users with `creator_oid` listed in `creators`, produces a record
    when any of them starts a livestream. When `creators` list is not provided,
    every livestream on the site will generate a record. `creator_oid` is the
    unique part of the user's home or livestream url.
    For example, `creator_oid` is `6596e71c04a7ea2fd7c36ae7`
    for the following urls:

    - `https://rplay.live/creatorhome/6596e71c04a7ea2fd7c36ae7`
    - `https://rplay.live/live/6596e71c04a7ea2fd7c36ae7`

    This monitor checks all creators with a single request,
    however, because of that it does not try to retrieve the
    direct `playlist url` for livestreams, and therefore does
    not support providing login credentials.
    """

    def __init__(self, conf: RplayMonitorConfig, entities: Sequence[RplayMonitorEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)
        self.nickname_cache: Dict[str, str] = {}
        self.consumers: Dict[str, asyncio.Queue] = {}
        for entity in self.entities.values():
            self.consumers[entity.name] = asyncio.Queue()

        # new data are only fetched in a single entity with the shortest update interval
        # it allows to avoid creating and monitoring a dedicated task for performing update
        entity_for_update = min(entities, key=lambda entity: entity.update_interval)
        entity_for_update.selected_for_update = True

    async def get_records(self, entity: RplayMonitorEntity, session: aiohttp.ClientSession) -> Sequence[RplayRecord]:
        if entity.selected_for_update:
            await self.update_records(entity, session)

        records = []
        queue = self.consumers[entity.name]
        while not queue.empty():
            records.append(queue.get_nowait())
        return records

    async def update_records(self, entity: RplayMonitorEntity, session: aiohttp.ClientSession) -> None:
        r = RplayUrl.livestreams()
        data = await self.request_json(r.url, entity, session, headers=r.headers, params=r.params)
        if data is None:
            return
        records = self.parse_livestreams(data)

        # nickname field in data from /livestreams endpoint is empty,
        # get them from /bulkgetusers instead
        await self.update_nicknames_cache(session, entity, records)
        self.update_nicknames(records)

        for record in records:
            for name, queue in self.consumers.items():
                creators = self.entities[name].creators
                if creators is None or record.creator_id in creators:
                    await queue.put(record)

    def parse_livestreams(self, data: List[dict]):
        if not isinstance(data, List):
            self.logger.warning(
                f'unexpected response from /livestreams endpoint, not a list of records. Raw response: {data}')
            return []
        records = []
        for live in data:
            try:
                record = parse_livestream(live)
                records.append(record)
            except Exception as e:
                self.logger.warning(f'failed to parse record: {type(e)} {e}. Raw data: "{data}"')
        return records

    async def update_nicknames_cache(self, session: aiohttp.ClientSession, entity: RplayMonitorEntity,
                                     records: List[RplayRecord]):
        creator_oids = [r.creator_id for r in records if r.creator_id not in self.nickname_cache]
        if not creator_oids:
            return
        r = RplayUrl.bulkgetusers(creator_oids)
        data = await utils.request_json(r.url, session, self.logger, headers=r.headers, params=r.params)
        if data is None:
            self.logger.warning(f'[{entity.name}] failed to update nickname cache: failed to get users info')
            return
        try:
            for oid, info in data.items():
                if 'nickname' in info:
                    self.nickname_cache[oid] = info['nickname']
        except Exception as e:
            self.logger.warning(f'[{entity.name}] failed to update nickname cache: {type(e)} {e}')

    def update_nicknames(self, records: List[RplayRecord]):
        for record in records:
            if record.creator_id in self.nickname_cache:
                record.name = self.nickname_cache[record.creator_id]


@Plugins.register('rplay.user', Plugins.kind.ACTOR_CONFIG)
class RplayUserMonitorConfig(BaseFeedMonitorConfig):
    login: Optional[str] = None
    """rplay login of the account used for monitoring"""
    password: Optional[str] = None
    """rplay password of the account used for monitoring"""
    user_id: Optional[str] = None
    """userOid of the account used for monitoring. When absent can be retrieved automatically by performing a login request"""

    @model_validator(mode='after')
    def check_invariants(self):
        if (self.login and not self.password) or (not self.login and self.password):
            raise ValueError('provide either both login and password or none of them')
        if self.user_id and not self.login:
            raise ValueError('when user_id is set,  credentials must also be provided')
        return self


@Plugins.register('rplay.user', Plugins.kind.ACTOR_ENTITY)
class RplayUserMonitorEntity(RplayMonitorEntity):
    creator_oid: str
    """ID of the user to monitor"""


@Plugins.register('rplay.user', Plugins.kind.ACTOR)
class RplayUserMonitor(BaseFeedMonitor):
    """
    Monitor livestreams on RPLAY channel

    Monitors a user with given `creator_oid`, produces record
    when the user starts a livestream. `creator_oid` is the
    unique part of the user's home or livestream url.
    For example, `creator_oid` is `6596e71c04a7ea2fd7c36ae7`
    for the following urls:

    - `https://rplay.live/creatorhome/6596e71c04a7ea2fd7c36ae7`
    - `https://rplay.live/live/6596e71c04a7ea2fd7c36ae7`

    When producing a record, this plugin will generate `playlist_url`
    for the stream. If credentials are provided in the `config` section,
    it will try to update it with the key required to access livestreams
    limited to subscribers. Resulting `playlist_url` might still be
    invalid if the update failed or account does not have permissions
    to view the stream.
    """

    def __init__(self, conf: RplayUserMonitorConfig, entities: Sequence[RplayUserMonitorEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)
        self.own_user: Optional['User'] = None

    async def get_records(self, entity: RplayUserMonitorEntity, session: aiohttp.ClientSession) -> Sequence[RplayRecord]:
        r = RplayUrl.play(entity.creator_oid)
        data = await self.request_json(r.url, entity, session, headers=r.headers, params=r.params)
        if data is None:
            return []
        try:
            record = parse_play(data)
        except Exception as e:
            self.logger.warning(f'failed to parse record: {type(e)} {e}. Raw data: "{data}"')
            return []
        if record is None:
            return []
        record.playlist_url = await self.get_playlist_url(session, record)
        return [record]

    async def get_playlist_url(self, session: aiohttp.ClientSession, record: RplayRecord) -> Optional[str]:
        if record.restream_url is not None:
            return None
        key2 = await self.get_key2(session) or ''
        r = RplayUrl.playlist(record.creator_id, key2=key2)
        return r.url_with_query

    async def get_key2(self, session: aiohttp.ClientSession) -> Optional[str]:
        own_user = await self.get_own_user(session)
        if own_user is None:
            return None
        r = RplayUrl.key2(own_user)
        raw_key2 = await utils.request(r.url, session, self.logger, r.method, r.params, r.data, r.headers, retry_times=3, retry_multiplier=3)
        if raw_key2 is None:
            return None
        try:
            key2 = json.loads(raw_key2)['authKey']
        except (TypeError, KeyError, json.JSONDecodeError) as e:
            self.logger.warning(f'failed to process key2 "{raw_key2}": {type(e)}:{e}')
            return None
        return key2

    async def get_own_user(self, session: aiohttp.ClientSession) -> Optional['User']:
        if self.conf.login is None:
            return None
        if self.own_user is None:
            token = User.token_for(self.conf.login, self.conf.password)
            if self.conf.user_id is not None:
                self.logger.debug(f'using userOid from configuration')
                user_oid = self.conf.user_id
            else:
                r = RplayUrl.login(token)
                user_info = await utils.request_json(r.url, session, self.logger, r.method, r.params, r.data, r.headers)
                if user_info is None:
                    self.logger.debug(f'failed to log in')
                    return None
                if not 'oid' in user_info:
                    self.logger.debug(f'user oid is absent in login response. Raw response: {user_info}')
                    return None
                user_oid = user_info['oid']
                self.logger.info(f'successfully logged in, user_id is "{user_oid}"')
            self.own_user = User(user_oid, token)
        return self.own_user


def parse_livestream(item: dict) -> RplayRecord:
    """parse a single item from a list returned by /live/livestreams endpoint"""
    creator_oid = item['creatorOid']

    restream_platform = item.get('streamState')
    restream_key = item.get('multiPlatformKey')
    restream_url = get_restream_url(restream_platform, restream_key)

    record = RplayRecord(

        url=get_livestream_url(creator_oid),
        title=item['title'],
        description=item['description'],
        thumbnail_url=live_thumbnail_url(creator_oid),
        start=dateutil.parser.parse(item['streamStartTime']),
        user_id=item['_id'],
        creator_id=creator_oid,
        name=item['creatorNickname'],
        avatar_url=get_avatar_url(creator_oid),
        restream_platform=restream_platform,
        restream_url=restream_url
    )
    return record


class StreamState(str, enum.Enum):
    OFFLINE = 'offline'
    LIVE = 'live'
    TWITCH = 'twitch'
    YOUTUBE = 'youtube'


def parse_play(item: dict) -> Optional[RplayRecord]:
    """parse a single item from a list returned by /live/play/ endpoint"""
    state = item['streamState']
    if state == StreamState.OFFLINE:
        return None
    elif state == StreamState.LIVE:
        restream_platform = None
        restream_key = None
    elif state == StreamState.TWITCH:
        restream_platform = state
        restream_key = item['twitchLogin']
    elif state == StreamState.YOUTUBE:
        restream_platform = state
        restream_key = item['liveStreamId']
    else:
        raise ValueError(f'unexpected stream state: {state}')
    restream_url = get_restream_url(restream_platform, restream_key)

    creator_oid = item['creatorOid']
    record = RplayRecord(
        url=get_livestream_url(creator_oid),
        title=item['title'],
        description=item['description'],
        thumbnail_url=live_thumbnail_url(creator_oid),
        start=dateutil.parser.parse(item['streamStartTime']),
        user_id=item['_id'],
        creator_id=creator_oid,
        name=item['creatorMetadata']['nickname'],
        avatar_url=get_avatar_url(creator_oid),
        restream_platform=restream_platform,
        restream_url=restream_url
    )
    return record


def get_avatar_url(creator_oid: str) -> str:
    return f'https://pb.rplay.live/profilePhoto/{creator_oid}'


def get_livestream_url(creator_oid: str) -> str:
    return f'https://rplay.live/live/{creator_oid}/'


def live_thumbnail_url(creator_oid: str) -> str:
    return f'https://pb.rplay.live/liveChannelThumbnails/{creator_oid}'


def get_restream_url(platform: Optional[str], restream_key: Optional[str]) -> Optional[str]:
    if platform is None or restream_key is None:
        return None
    if platform == 'twitch':
        return f'https://twitch.tv/{restream_key}'
    if platform == 'youtube':
        return f'https://www.youtube.com/watch?v={restream_key}'
    return None


class JWTAuth:

    @staticmethod
    def _dict_to_b64url(data: dict) -> str:
        data_json = json.dumps(data).replace(' ', '')
        data_b64url = base64.urlsafe_b64encode(data_json.encode())
        return data_b64url.decode().strip('=')

    @staticmethod
    def _hmac_signature_b64url(message: str, secret: str) -> str:
        signature = hmac.new(secret.encode(), msg=message.encode(), digestmod='sha256').digest()
        return base64.urlsafe_b64encode(signature).decode().strip('=')

    @classmethod
    def generate_token(cls, eml: str, dat: Optional[datetime.datetime] = None) -> str:
        header = {'alg': 'HS256', 'typ': 'JWT'}
        header_encoded = cls._dict_to_b64url(header)

        dat = dat or datetime.datetime.now()
        dat = dat.astimezone()
        dat_text = dat.strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z'
        iat = int(dat.timestamp())
        if dat.utcoffset() is not None:
            # timestamp is calculated from local time as if it was UTC
            iat += int(dat.utcoffset().total_seconds())

        payload = {'eml': eml, 'dat': dat_text, 'iat': iat}
        payload_encoded = cls._dict_to_b64url(payload)

        return f'{header_encoded}.{payload_encoded}'

    @classmethod
    def sign_token(cls, token: str, secret: str) -> str:
        secret = hashlib.sha256(secret.encode()).digest().hex()
        signature = cls._hmac_signature_b64url(token, secret)
        return f'{token}.{signature}'

    @classmethod
    def new(cls, eml: str, psw: str, dat: Optional[datetime.datetime] = None) -> str:
        """generate new signed token"""
        token = cls.generate_token(eml, dat)
        signed_token = cls.sign_token(token, psw)
        return signed_token


@dataclass
class User:
    user_oid: str
    token: str

    @staticmethod
    def token_for(email: str, passwd: str) -> str:
        return JWTAuth.new(email, passwd)

    def get_auth_header(self) -> Dict[str, str]:
        return {'Authorization': self.token}


@dataclass
class RequestDetails:
    url: str
    method: str = 'GET'
    params: Optional[Dict[str, Any]] = None
    data: Optional[Any] = None
    headers: Optional[Dict[str, Any]] = None

    @property
    def url_with_query(self) -> str:
        if self.params is None:
            return self.url
        parsed = urllib.parse.urlparse(self.url)
        with_query = parsed._replace(query=urllib.parse.urlencode(self.params))
        url = urllib.parse.urlunparse(with_query)
        return url


class RplayUrl:

    @staticmethod
    def livestreams(creator_oid: str = '') -> RequestDetails:
        url = f'https://api.rplay.live/live/livestreams'
        params = {'creatorOid': creator_oid, 'lang': 'en'}
        return RequestDetails(url=url, params=params)

    @staticmethod
    def play(oid: str, key: str = '') -> RequestDetails:
        """
        'streamState': 'offline', 'live' | 'twitch' | 'youtube'
        'liveStreamId' - youtube video_id or ""
        'twitchLogin' - twitch username or ""
        'streamStartTime': '2024-07-23T11:29:29.000Z'
        '_id', 'creatorOid', 'title', 'description'
        'creatorMetadata': {
            'nickname',
            'channelImage' - channel banner url,
            'customUrl' - custom channel name, "/c:creator",
            'published', 'publishedClips', 'playlists', 'communityPosts', 'pinnedPost' - list of contentOid
        }
        """
        url = f'https://api.rplay.live/live/play'
        params = {'creatorOid': oid, 'key': key, 'lang': 'en'}
        return RequestDetails(url=url, params=params)

    @staticmethod
    def getuser_by_oid(creator_oid: str) -> RequestDetails:
        """get user info and live status by userOid"""
        url = f'https://api.rplay.live/account/getuser'
        params = {'userOid': creator_oid, 'filter[]': ['_id', 'nickname', 'creatorTags'], 'lang': 'en'}
        return RequestDetails(url=url, params=params)

    @staticmethod
    def getuser_by_name(cursom_url: str) -> RequestDetails:
        """get user info and live status by custom channel url"""
        url = f'https://api.rplay.live/account/getuser'
        params = {'customUrl': cursom_url, 'filter[]': ['_id', 'nickname', 'creatorTags'], 'lang': 'en'}
        return RequestDetails(url=url, params=params)

    @staticmethod
    def subscriptions(oid: str) -> RequestDetails:
        url = f'https://api.rplay.live/account/getuser'
        params = {'userOid': oid, 'filter[]': ['_id', 'nickname', 'subscribingTo'], 'lang': 'en'}
        return RequestDetails(url=url, params=params)

    @staticmethod
    def bulkgetusers(oids: List[str]) -> RequestDetails:
        url = f'https://api.rplay.live/account/bulkgetusers'
        params = {'users': '|'.join(oids),
                  'toGrab': '|'.join(['_id', 'nickname', 'lastPubDate', 'creatorTags', 'isLive']), 'lang': 'en'}
        return RequestDetails(url=url, params=params)

    @staticmethod
    def content(content_oid, user: Optional[User] = None) -> RequestDetails:
        url = f'https://api.rplay.live/content'
        params = {'contentOid': content_oid, 'status': 'published', 'withContentMetadata': True, 'requestCanView': True, 'lang': 'en'}
        headers = {}
        if user is not None:
            params.update({'requestorOid': user.user_oid, 'loginType': 'plax'})
            headers.update(user.get_auth_header())
        return RequestDetails(url=url, params=params, headers=headers)

    @staticmethod
    def playlist(creator_oid: str, key: str = '', key2: str = '') -> RequestDetails:
        url = 'https://api.rplay.live/live/stream/playlist.m3u8'
        params = {'creatorOid': creator_oid, 'key': key, 'key2': key2}
        return RequestDetails(url=url, params=params)

    @staticmethod
    def key2(user: User) -> RequestDetails:
        """response is the key as a plaintext"""
        url = f'https://api.rplay.live/live/key2'
        params = {'requestorOid': user.user_oid, 'loginType': 'plax'}
        headers = user.get_auth_header()
        return RequestDetails(url=url, params=params, headers=headers)

    @staticmethod
    def login(token: str) -> RequestDetails:
        url = f'https://api.rplay.live/account/login'
        data = {'checkAdmin': None, 'lang': 'en', 'loginType': None, 'token': token}
        headers = {'Content-Type': 'application/json'}
        return RequestDetails(url=url, method='POST', data=json.dumps(data), headers=headers)



#!/usr/bin/env python3

import argparse
import asyncio
import logging
from asyncio import AbstractEventLoop
from pathlib import Path
from typing import Any, Dict, Optional, Tuple

from avtdl.core import webui
from avtdl.core.chain import Chain
from avtdl.core.config import ConfigParser, ConfigurationError, SettingsSection, config_sancheck
from avtdl.core.info import generate_plugins_description, generate_version_string
from avtdl.core.interfaces import Actor, RuntimeContext, TerminatedAction
from avtdl.core.loggers import set_logging_format, silence_library_loggers
from avtdl.core.plugins import UnknownPluginError
from avtdl.core.utils import read_file, write_file
from avtdl.core.yaml import yaml_load

DEFAULT_CONFIG_PATH = Path('config.yml')
CONFIG_TEMPLATE = '''
actors:
  rss:
    entities:
      - name: "ChannelName"
        url: "https://www.youtube.com/feeds/videos.xml?channel_id=UCK0V3b23uJyU4N8eR_BR0QA"
      - name: "AnotherChannelName"
        url: "https://www.youtube.com/feeds/videos.xml?channel_id=UC3In1x9H3jC4JzsIaocebWg"
  execute:
    entities:
      - name: "archive"
        command: "ytarchive --threads 3 --wait {url} best"
        working_dir: "archive/livestreams/{author}/"

chains:
  "archive":
    - rss:
        - "ChannelName"
        - "AnotherChannelName"
    - execute:
        - "archive"
'''


def load_config(path: Path) -> Any:
    try:
        if not path.exists():
            alt_path = path.with_suffix(path.suffix + '.txt')
            if alt_path.exists():
                print(f'Configuration file {path} not found, trying {alt_path} instead.')
                path = alt_path
            elif path == DEFAULT_CONFIG_PATH:
                print(f'Configuration file {path} not found, using example template.')
                write_file(path, CONFIG_TEMPLATE)
            else:
                raise ValueError('Configuration file {} does not exist'.format(path))
        config_text = read_file(path)
        config = yaml_load(config_text)
    except Exception as e:
        print('Failed to parse configuration file:')
        print(e)
        raise SystemExit from e
    return config


def parse_config(conf, ctx: Optional[RuntimeContext] = None) -> Tuple[SettingsSection, Dict[str, Actor], Dict[str, Chain]]:
    ctx = ctx or RuntimeContext.create()
    try:
        settings, actors, chains = ConfigParser.parse(conf, ctx)
    except (ConfigurationError, UnknownPluginError) as e:
        logging.error(e)
        raise SystemExit from e
    except Exception as e:
        logging.exception(e)
        raise SystemExit from e
    return settings, actors, chains


def handler(loop: AbstractEventLoop, context: Dict[str, Any]) -> None:
    logging.exception(f'unhandled exception in event loop:', exc_info=context.get('exception'))
    loop.default_exception_handler(context)


async def install_exception_handler() -> None:
    loop = asyncio.get_running_loop()
    loop.set_exception_handler(handler)
    loop.slow_callback_duration = 100



async def run(config_path: Path, host: Optional[str], port: Optional[int]) -> None:
    await install_exception_handler()
    while True:
        config = load_config(config_path)
        ctx = RuntimeContext.create()
        with ctx:
            settings, actors, chains = parse_config(config, ctx)
            config_sancheck(actors, chains)

            if host is not None:
                settings.host = host
            if port is not None:
                settings.port = port

            controller = ctx.controller
            for runnable in actors.values():
                _ = controller.create_task(runnable.run(), name=f'{runnable!r}.{hash(runnable)}')
            _ = controller.create_task(webui.run(config_path, config, ctx, settings, actors, chains), name='webui')

            action = await controller.run_until_termination()
            if action == TerminatedAction.EXIT:
                logging.info('terminating...')
                break
            elif action == TerminatedAction.RESTART:
                logging.info('restarting...')
                continue
            else:
                assert False, f'Unknown action: {action}'


def make_docs(output: Path) -> None:
    doc = generate_plugins_description()
    try:
        with open(output, 'wt', encoding='utf8') as fp:
            fp.write(doc)
    except OSError as e:
        logging.error(f'failed to write documentation file "{output}": {e}')
        raise SystemExit from e


def main() -> None:
    description = '''Tool for monitoring rss feeds and other sources and running commands for new entries'''
    parser = argparse.ArgumentParser(description=description)
    help_d = 'set loglevel to DEBUG'
    parser.add_argument('-d', '--debug', action='store_true', default=False, help=help_d)
    help_v = 'print version and exit'
    parser.add_argument('-v', '--version', action='store_true', default=False, help=help_v)
    help_c = 'specify path to configuration file to use instead of default'
    parser.add_argument('-c', '--config', type=Path, default=DEFAULT_CONFIG_PATH, help=help_c)
    help_h = 'write plugins documentation in markdown format into a given file and exit'
    parser.add_argument('-p', '--plugins-doc', type=Path, required=False, help=help_h)
    help_dp = 'web-interface port (takes priority over configuration file)'
    parser.add_argument('-P', '--port', type=int, required=False, help=help_dp)
    help_dh = 'web-interface bind address (takes priority over configuration file)'
    parser.add_argument('-H', '--host', required=False, help=help_dh)
    args = parser.parse_args()

    if args.debug:
        log_level = logging.DEBUG
    else:
        log_level = logging.INFO

    set_logging_format(log_level)
    silence_library_loggers()

    try:
        if args.version:
            print(generate_version_string())
        elif args.plugins_doc is not None:
            make_docs(args.plugins_doc)
        else:
            asyncio.run(run(args.config, args.host, args.port), debug=True)
    except KeyboardInterrupt:
        if args.debug:
            logging.exception('Interrupted, exiting... Printing stacktrace for debugging purpose:')
        else:
            logging.info('Interrupted, exiting...')


if __name__ == "__main__":
    main()



import logging
import os
import re
from pathlib import Path
from typing import List

import pytest

from avtdl.avtdl import load_config, parse_config
from avtdl.core.config import config_sancheck
from avtdl.core.loggers import silence_library_loggers


def read_file(path: str) -> str:
    with open(path, 'rt', encoding='utf8') as fp:
        return fp.read()


def text_as_path(directory, text: str) -> Path:
    directory = Path(directory)
    filename = directory / str(hash(text))
    with filename.open('wt', encoding='utf8') as fp:
        fp.write(text)
    return filename


def find_examples(text) -> List[str]:
    """return list with code blocks in yaml format"""
    return re.findall('```yaml([^`]+)```', text)


def configs(file: str) -> List[str]:
    configs_text = read_file(file)
    example_configs = find_examples(configs_text)
    return example_configs


def empty_cookies(tmp_path: Path):
    text = '# Netscape HTTP Cookie File\n\n'
    with open(tmp_path / 'cookies.txt', 'wt', encoding='utf8') as fp:
        fp.write(text)


# fixtures does not exist yet when parametrization happens,
# therefore it is not possible to use return value of a fixture
# for tests parametrization. All preparations are done by regular
# functions instead
examples = [*configs('EXAMPLES.md'), *configs('example.config.yml')]
ids = [f'example_{i}' for i in range(len(examples))]


@pytest.mark.parametrize('config', examples, ids=ids)
def test_config_loading(config: str, tmp_path, caplog):
    """
    Smoke test for configuration examples

    Extracts code blocks containing yaml from examples docs
    and tries parsing and loading each of them as config.

    Value of the cookies_file parameter used in examples
    must always be 'cookies.txt'.
    """

    silence_library_loggers()
    caplog.set_level(logging.WARNING)

    os.chdir(tmp_path)
    empty_cookies(tmp_path)
    path = text_as_path(tmp_path, config)

    conf = load_config(path)
    _, actors, chains = parse_config(conf)
    config_sancheck(actors, chains)

    assert not caplog.records



from io import StringIO
from typing import Any, Union

from ruamel.yaml import CommentedMap, CommentedSeq, YAML

CommentedData = Union[CommentedMap, CommentedSeq, Any]
PlainData = Union[dict, list, Any]


def yaml_load(text: str) -> CommentedData:
    return make_yaml_parser().load(text)


def yaml_dump(data: PlainData) -> str:
    stream = StringIO()
    make_yaml_parser().dump(data, stream)
    return stream.getvalue()


def make_yaml_parser() -> YAML:
    yaml_parser = YAML()
    yaml_parser.indent(mapping=2, sequence=4, offset=2)
    yaml_parser.preserve_quotes = True
    return yaml_parser


def merge_data(base: CommentedData, data: PlainData) -> CommentedData:
    if str(base) == str(data):
        return base
    elif isinstance(base, CommentedSeq) and isinstance(data, list):
        merge_seq(base, data)
        return base
    elif isinstance(base, CommentedMap) and isinstance(data, dict):
        merge_map(base, data)
        return base
    else:
        return data


def merge_map(base: CommentedMap, data: dict) -> None:
    to_delete = [k for k in base if k not in data]
    for k in to_delete:
        del base[k]
    for i, (k, v) in enumerate(data.items()):
        if k in base:
            base[k] = merge_data(base[k], v)
        else:
            base.insert(i, k, v)


def merge_seq(base: CommentedSeq, data: list) -> None:
    old_base = {str(v): v for v in base}
    for v in old_base.values():
        base.remove(v)
    for v in data:
        if str(v) in old_base:
            value = old_base[str(v)]
        else:
            value = v
        base.append(value)



import datetime
from typing import Optional

import pytest

from avtdl.core.interfaces import RuntimeContext, TextRecord
from avtdl.plugins.filters.filters import EmptyFilterConfig, FormatFilter, FormatFilterEntity


class TestFormatFilter:

    @staticmethod
    @pytest.fixture()
    def text_record():
        return TextRecord(text='test text message')

    @staticmethod
    def prepare_filter(template: str, missing: Optional[str] = None):
        config = EmptyFilterConfig(name='test')
        entity = FormatFilterEntity(name='test', template=template, missing=missing)
        ctx = RuntimeContext.create()
        return FormatFilter(config, [entity], ctx), entity

    def test_single_field(self, text_record):
        fmt, entity = self.prepare_filter('*** {text} ***')

        result = fmt.match(entity, text_record)

        assert result.text == '*** test text message ***'

    def test_missing(self, text_record):
        fmt, entity = self.prepare_filter('*** {url} ***', missing='---')

        result = fmt.match(entity, text_record)

        assert result.text == '*** --- ***'

    def test_missing_disabled(self, text_record):
        fmt, entity = self.prepare_filter('*** {url} ***')

        result = fmt.match(entity, text_record)

        assert result.text == '*** {url} ***'

    def test_empty_placeholder(self, text_record):
        fmt, entity = self.prepare_filter('*** {} ***')

        result = fmt.match(entity, text_record)

        assert result.text == '*** {} ***'

    def test_nested_placeholder(self, text_record):
        fmt, entity = self.prepare_filter('*** {{text}} ***')

        result = fmt.match(entity, text_record)

        assert result.text == '*** {test text message} ***'

    def test_escaped_placeholder(self, text_record):
        fmt, entity = self.prepare_filter('*** \{text\}=\{{text}\} ***')

        result = fmt.match(entity, text_record)

        assert result.text == '*** {text}={test text message} ***'

    def test_unpaired_braces(self, text_record):
        fmt, entity = self.prepare_filter('}{ {text} }{')

        result = fmt.match(entity, text_record)

        assert result.text == '}{ test text message }{'

    def test_datetime(self, text_record):
        fmt, entity = self.prepare_filter('[%Y-%m-%d] {text}')
        now = datetime.datetime.now()
        result = fmt.match(entity, text_record)

        assert result.text == f'[{now.year}-{now.month:02}-{now.day:02}] test text message'



import datetime
import json
import logging
import re
from textwrap import shorten
from typing import Any, Dict, List, Optional, Tuple, Union

import dateutil.parser
from pydantic import BaseModel, ValidationError

from avtdl.core.config import format_validation_error
from avtdl.core.interfaces import MAX_REPR_LEN, Record
from avtdl.core.utils import Fmt, find_one

local_logger = logging.getLogger().getChild('twitter_extractors')


class TwitterRecord(Record):
    """
    Single post as parsed from Twitter instance

    Depending on the tweet type (regular, retweet, reply, quote) some fields might be empty
    """
    uid: str
    """tweet id"""
    url: str
    """tweet url"""
    author: str
    """user's visible name"""
    username: str
    """user's handle"""
    avatar_url: Optional[str] = None
    """link to the picture used as the user's avatar"""
    published: datetime.datetime
    """tweet timestamp"""
    text: str
    """tweet text. Links are unshortened if possible"""
    attachments: List[str] = []
    """list of links to attached images or video thumbnails"""
    images: List[str] = []
    """list of links to attached images"""
    videos: List[str] = []
    """list of links to attached videos and gifs"""
    replying_to_username: Optional[str] = None
    """for replies, name of the user that got replied to"""
    retweet: Optional['TwitterRecord'] = None
    """for retweets, nested TwitterRecord containing tweet that was retweeted"""
    quote: Optional['TwitterRecord'] = None
    """for quotes, nested TwitterRecord containing tweet being quoted"""
    space_id: Optional[str] = None
    """for tweets or retweets mentioning Twitter Space url contains unique part of the url, otherwise empty"""

    def model_post_init(self, __context: Any) -> None:
        self.space_id = find_space_id(self.text)

    def __repr__(self):
        return f'TwitterRecord(author="{self.author}", url="{self.url}", text="{shorten(self.text, MAX_REPR_LEN)}")'

    def __str__(self):
        tweet = self.retweet or self
        elements = []
        elements.append(tweet.url)
        if self.retweet is not None:
            retweet_header = f'[{self.published.strftime("%Y-%m-%d %H:%M:%S")}] {self.author} (@{self.username}) has retweeted:'
            elements.append(retweet_header)
        if tweet.replying_to_username is not None:
            reply_header = f'Replying to @{tweet.replying_to_username}:'
            elements.append(reply_header)
        elements.append(f'[{tweet.published.strftime("%Y-%m-%d %H:%M:%S")}] {tweet.author} (@{tweet.username}):')
        elements.append(tweet.text)
        if tweet.attachments:
            elements.append('\n'.join(tweet.attachments))
        if tweet.quote:
            elements.append('\nReferring to ')
            elements.append(str(tweet.quote))
        return '\n'.join(elements)

    def get_uid(self) -> str:
        return self.url

    def discord_embed(self) -> List[dict]:
        quote = self.quote if self.retweet is None else self.retweet.quote

        text_items = [self.text]
        if len(self.attachments) > 1:
            text_items.extend(self.attachments)
        if quote:
            text_items.append('\nReferring to ')
            text_items.append(str(quote))
        text = '\n'.join(text_items)

        if self.retweet is not None:
            author = f'[{self.published.strftime("%Y-%m-%d %H:%M:%S")}] {self.author} (@{self.username}) has retweeted:'
            title = f'{self.retweet.author} ({self.retweet.username})'
            if self.replying_to_username:
                title += f', replying to @{self.replying_to_username}:'
            avatar = self.retweet.avatar_url
            timestamp = self.retweet.published.isoformat()
        else:
            author = f'{self.author} ({self.username})'
            title = f'Replying to @{self.replying_to_username}:' if self.replying_to_username else self.url
            avatar = self.avatar_url
            timestamp = self.published.isoformat()

        embed = {
            'title': title,
            'description': text,
            'url': self.url,
            'color': None,
            'author': {'name': author, 'icon_url': avatar, 'url': user_url_by_id(self.username)},
            'timestamp': timestamp,
        }

        def format_attachments(post_url: str, attachments: List[str]) -> List[dict]:
            return [{'url': post_url, 'image': {'url': attachment}} for attachment in attachments]

        if self.attachments:
            images = format_attachments(self.url, self.attachments)
            embed['image'] = images.pop(0)['image']
            embeds = [embed, *images]
        elif quote and quote.attachments:
            images = format_attachments(quote.url, quote.attachments)
            embed['image'] = images.pop(0)['image']
            embeds = [embed, *images]
        else:
            embeds = [embed]
        return embeds


def find_space_id(text: str) -> Optional[str]:
    rest_id_match = re.search(r'/i/spaces/([0-9a-zA-Z]+)/?', text)
    if rest_id_match is None:
        return None
    return rest_id_match.groups()[0]


class UserInfo(BaseModel):
    rest_id: str
    handle: str
    name: str
    description: str
    avatar_url: str
    banner_url: Optional[str] = None
    location: str

    @classmethod
    def from_data(cls, data: dict) -> 'UserInfo':
        result = find_one(data, '$.data.user.result')
        if result is None:
            raise ValueError(f'failed to parse data into {cls.__name__}: no "result" property')
        return cls.from_result(result)

    @classmethod
    def from_result(cls, result: dict) -> 'UserInfo':
        typename = result.get('__typename')
        if typename != 'User':
            raise ValueError(f'failed to parse result into {cls.__name__}: __typename is "{typename}, expected "User"')
        rest_id = result['rest_id']

        legacy = result['legacy']

        handle = legacy['screen_name']
        name = legacy['name']
        description = legacy['description']
        avatar_url = legacy['profile_image_url_https'].replace('_normal', '')
        banner_url = legacy.get('profile_banner_url')
        location = legacy['location']
        return cls(rest_id=rest_id, handle=handle, name=name, description=description, avatar_url=avatar_url, banner_url=banner_url, location=location)


def extract_contents(data: str) -> Tuple[List[dict], Optional[str]]:
    """Picks all tweets, individual and inside conversations. Also picks bottom cursor value"""
    tweets: List[dict] = []
    continuation = None

    def handle_item(obj):
        nonlocal continuation
        # drop pinned tweet
        if "type" in obj and obj['type'] == 'TimelinePinEntry':
            try:
                pinned = obj['entry']['content']['itemContent']['tweet_results']
            except IndexError:
                local_logger.warning(f'TimelinePinEntry is present but no "tweet_result" was found inside. Raw data:\n"{obj}"')
                return obj
            if tweets and tweets[-1] == pinned:
                # when handle_item() gets called for "TimelinePinEntry", it must have been
                # already called for the "tweet_results" inside it, which then should be
                # on top of the "tweets" list now
                tweets.pop()
            else:
                local_logger.warning(f'TimelinePinEntry is present but pinned tweet was not collected. Raw data:\n"{obj}"')
            return obj
        if '__typename' not in obj:
            return obj
        # tweet
        if 'tweet_results' in obj:
            tweets.append(obj['tweet_results'])
            return obj
        # cursor
        typename = obj['__typename']
        if typename == 'TimelineTimelineCursor':
            if obj.get('cursorType') == 'Bottom':
                continuation = obj.get('value')
            return obj
        return obj

    decoder = json.JSONDecoder(object_hook=handle_item)
    data, _ = decoder.raw_decode(data)
    return tweets, continuation


def parse_tweet(tweet_results: dict) -> TwitterRecord:
    tweet_result = tweet_results.get('result')
    if tweet_result is None:
        raise ValueError(f'failed to parse tweet: no "result"')
    typename = tweet_result.get('__typename')
    if typename == 'TweetWithVisibilityResults':
        tweet_result = tweet_result['tweet']
    elif typename != 'Tweet':
        raise ValueError(f'failed to parse tweet: __typename is "{typename}, expected "Tweet"')
    rest_id = tweet_result['rest_id']

    try:
        user_result = tweet_result['core']['user_results']['result']
    except (KeyError, TypeError):
        raise ValueError(f'failed to parse tweet: no user_result found')
    try:
        user = UserInfo.from_result(user_result)
    except Exception as e:
        raise ValueError(f'failed to parse tweet: {e}')
    url = tweet_url_by_id(user.handle, rest_id)

    legacy = tweet_result.get('legacy')
    if legacy is None:
        raise ValueError(f'failed to parse tweet: no "legacy" field in tweet_result for {url}')

    published = tweet_timestamp(rest_id) or dateutil.parser.parse(legacy['created_at'])

    try:
        text = note_text(tweet_result) or tweet_text(tweet_result)
    except Exception as e:
        raise ValueError(f'failed to parse tweet text: {e}')

    attachments, images, videos = parse_media(tweet_result) or []

    replying_to_username: Optional[str] = legacy.get('in_reply_to_screen_name')

    retweet_result = legacy.get('retweeted_status_result')
    retweet = parse_tweet(retweet_result) if retweet_result else None
    if retweet is not None:
        text = retweet.text

    quote_tesults = tweet_result.get('quoted_status_result') or {}
    quote_tesult = quote_tesults.get('result') or {}
    if quote_tesult.get('__typename') == 'Tweet':
        quote: Optional[TwitterRecord] = parse_tweet(quote_tesults)
    elif quote_tesult.get('__typename') == 'TweetTombstone':
        quote = parse_quoted_tombstone(tweet_result)
    elif legacy.get('quoted_status_id_str') and not retweet:
        # 'quoted_status_result' might be empty despite quote fields being present in 'legacy'
        quote = parse_quoted_tombstone(tweet_result)
    else:
        quote = None

    tweet = TwitterRecord(
        uid=rest_id,
        url=url,
        author=user.name,
        username=user.handle,
        avatar_url=user.avatar_url,
        published=published,
        text=text,
        attachments=attachments,
        images=images,
        videos=videos,
        replying_to_username=replying_to_username,
        retweet=retweet,
        quote=quote
    )
    return tweet


def parse_quoted_tombstone(tweet_result: dict) -> Optional[TwitterRecord]:
    """given valid tweet that quotes tombstone, try to extract some data on the tombstone"""
    try:
        text = tweet_result['quoted_status_result']['result']['tombstone']['text']['text']
    except (KeyError, TypeError):
        text = 'Tweet content is unavailable'

    try:
        legacy = tweet_result['legacy']
        url = legacy['quoted_status_permalink']['expanded']
        rest_id = legacy['quoted_status_id_str']
        published = tweet_timestamp(rest_id) or datetime.datetime.fromtimestamp(0, tz=datetime.timezone.utc)

        username_match = re.search(r'/([^/]+)/status/\d+', url)
        username = username_match.groups()[0] if username_match else 'username_missing'
    except Exception:
        return None

    return TwitterRecord(
        uid=rest_id,
        url=url,
        author=username,
        username=username,
        published=published,
        text=text)


def parse_media(tweet_result: dict) -> Tuple[List[str], List[str], List[str]]:
    try:
        legacy = tweet_result['legacy']
    except KeyError as e:
        raise ValueError(f'failed to parse tweet media: no {e} found')
    media = (legacy.get('entities') or {}).get('media', [])
    extended_media = (legacy.get('extended_entities') or {}).get('media', [])
    if not media and not extended_media:
        return [], [], []
    attachments = [media_item['media_url_https'] for media_item in media]
    images = []
    videos = []
    for item in media:
        item_type = item['type']
        if item_type == 'photo':
            media_url = item['media_url_https']
            images.append(media_url)
        elif item_type in ['video', 'animated_gif']:
            try:
                variants = item['video_info']['variants']
            except (KeyError, TypeError):
                variants = []
            best_variant = max(variants, key=lambda variant: variant.get('bitrate', -1))
            video_url = best_variant['url']
            videos.append(video_url)
        else:
            msg = f'unknown media type {item_type}. Raw tweet_result:\n"{tweet_result}"'
            local_logger.debug(msg)
    return attachments, images, videos


def replace_urls(text: str, urls: List[Dict[str, str]]) -> str:
    for url in urls:
        try:
            text = text.replace(url['url'], url['expanded_url'])
        except KeyError:
            pass
    return text


def tweet_text(tweet_result: dict) -> str:
    try:
        legacy = tweet_result['legacy']
        text = legacy['full_text']
        entities = legacy['entities']
    except (KeyError, TypeError) as e:
        raise ValueError(f'failed to parse tweet text: no {e} found')
    urls = entities.get('urls', [])
    text = replace_urls(text, urls)
    img_urls = entities.get('media', [])
    # remove media links from the bottom of the tweet
    for url in img_urls:
        try:
            text = text.replace(url['url'], '')
        except KeyError:
            pass
    return text


def note_text(tweet_result: dict) -> Optional[str]:
    note = tweet_result.get('note_tweet')
    if note is None:
        return None
    try:
        result = note['note_tweet_results']['result']
    except (KeyError, TypeError):
        local_logger.debug(f'failed to parse note_tweet: no result. Raw tweet_result: "{tweet_result}"')
        return None
    text = result.get('text')
    if text is None:
        local_logger.debug(f'failed to parse note_tweet: no text. Raw tweet_result: "{tweet_result}"')
        return None
    urls = result.get('entity_set', {}).get('urls', [])
    if urls:
        text = replace_urls(text, urls)
    return text


def maybe_date(timestamp: Union[int, str, None], tz: Optional[datetime.timezone] = datetime.timezone.utc) -> Optional[datetime.datetime]:
    if timestamp is None:
        return None
    try:
        timestamp = int(timestamp)
        date = datetime.datetime.fromtimestamp(timestamp / 1000, tz=tz)
        return date
    except Exception:
        return None


def tweet_timestamp(tweet_id: Union[str, int]) -> Optional[datetime.datetime]:
    try:
        tweet_id = int(tweet_id)
    except ValueError:
        return None
    if tweet_id < 30_000_000_000:  # old incremental id
        return None
    try:
        timestamp = ((tweet_id >> 22) + 1288834974657) / 1000
        return datetime.datetime.fromtimestamp(timestamp, tz=datetime.timezone.utc)
    except Exception:
        return None


def parse_timeline(text: str) -> Tuple[List[TwitterRecord], Optional[str]]:
    tweets = []
    raw_tweets, continuation = extract_contents(text)
    for tweet_result in raw_tweets:
        tweet = parse_tweet(tweet_result)
        tweets.append(tweet)
    return tweets, continuation


def space_url_by_id(space_id: str) -> str:
    return f'https://twitter.com/i/spaces/{space_id}'


def user_url_by_id(handle: str) -> str:
    return f'https://twitter.com/{handle}'


def tweet_url_by_id(handle: str, rest_id: str) -> str:
    return f'https://twitter.com/{handle}/status/{rest_id}'


def parse_space(data: dict) -> 'TwitterSpaceRecord':
    metadata = find_one(data, '$..metadata')
    if metadata is None:
        raise ValueError(f'failed to parse space: no metadata found')
    try:
        user_result = metadata['creator_results']['result']
        user = UserInfo.from_result(user_result)
    except (KeyError, TypeError):
        raise ValueError(f'failed to parse space: no creator_result found')
    except ValidationError as e:
        err = format_validation_error(e)
        raise ValueError(f'failed to parse space author details: {err}')
    try:
        uid = metadata.get('rest_id')
        record = TwitterSpaceRecord(
            uid=uid,
            url=space_url_by_id(uid),
            state=metadata.get('state'),
            media_key=metadata.get('media_key'),
            title=metadata.get('title', ''),
            author=user.name,
            username=user.handle,
            avatar_url=user.avatar_url,
            published=maybe_date(metadata.get('created_at')) or maybe_date(metadata.get('started_at')) or datetime.datetime.now(tz=datetime.timezone.utc),
            scheduled=maybe_date(metadata.get('scheduled_start')),
            started=maybe_date(metadata.get('started_at')),
            ended=maybe_date(metadata.get('ended_at')),
            updated=maybe_date(metadata.get('updated_at')),
            recording_enabled=metadata.get('is_space_available_for_replay', False)
        )
    except ValidationError as e:
        err = format_validation_error(e)
        raise ValueError(f'failed to parse space: {err}')
    return record


def parse_media_url(data: dict) -> Optional[str]:
    source = data.get('source') or {}
    media_url = source.get('location') or source.get('noRedirectPlaybackUrl') or None
    return media_url


class TwitterSpaceRecord(Record):
    uid: str
    """space id"""
    url: str
    """url of the space"""
    state: str
    """description of current status of a space: upcoming, ongoing, ended"""
    media_key: str
    """id that can be used to fetch url of the underlying HLS stream"""
    media_url: Optional[str] = None
    """link to the underlying stream, when available"""
    title: str
    """space title"""
    author: str
    """user's visible name"""
    username: str
    """user's handle"""
    avatar_url: Optional[str] = None
    """link to the picture used as the user's avatar"""
    published: datetime.datetime
    """timestamp of the space creation"""
    scheduled: Optional[datetime.datetime] = None
    """scheduled time for an upcoming space to start at, otherwise absent"""
    started: Optional[datetime.datetime] = None
    """timestamp of the space start, empty for upcoming spaces"""
    ended: Optional[datetime.datetime] = None
    """timestamp of the space end, empty for not yet ended spaces"""
    updated: Optional[datetime.datetime] = None
    """timestamp of the last update"""
    recording_enabled: bool = False
    """whether host enabled recording of the space. When true, archive is likely to be available"""

    def __str__(self) -> str:
        header = f'[{self.published.strftime("%Y-%m-%d %H:%M:%S")}] Twitter Space by {self.author} (@{self.username}) [{self.state}]'
        if not self.recording_enabled:
            header += ' [unarchived]'
        scheduled = f'\nscheduled at {self.scheduled}' if self.scheduled else ''
        return f'{self.url}\n{header}\n{self.title}{scheduled}'

    def __repr__(self):
        return f'TwitterSpaceRecord(author="{self.author}", url="{self.url}")'

    def get_uid(self) -> str:
        return self.uid

    def discord_embed(self) -> List[dict]:
        author = f'{self.author} ({self.username})'
        fields: List[dict] = []
        state_text = ' '.join(re.split('(?=[A-Z])', self.state)).strip().lower()
        fields.append({'name': f'[{state_text}]', 'value': '', 'inline': True})
        if not self.recording_enabled:
            fields.append({'name': '[unarchived]', 'value': '', 'inline': True})
        if self.scheduled:
            fields.append({'name': 'scheduled', 'value': Fmt.dtf(self.scheduled), 'inline': False})
        if self.started:
            fields.append({'name': 'started', 'value': Fmt.dtf(self.started), 'inline': False})
        if self.ended:
            fields.append({'name': 'ended', 'value': Fmt.dtf(self.ended), 'inline': False})

        embed = {
            'title': self.title or self.url,
            'description': self.media_url,
            'url': self.url,
            'color': None,
            'author': {'name': author, 'icon_url': self.avatar_url, 'url': user_url_by_id(self.username)},
            'timestamp': self.published.isoformat(),
            'fields': fields
        }

        return [embed]



import asyncio
import logging
import sys
from dataclasses import dataclass
from typing import Callable, Optional

import slixmpp

if sys.platform == 'win32':
    slixmpp.xmlstream.resolver.AIODNS_AVAILABLE = False

ON_ERROR_RETRY_DELAY = 60
DISCONNECT_AFTER_DONE_DELAY = 30


@dataclass
class Line:
    recipient: str
    message: str


class MSG2JBR:
    def __init__(self, username: str, passwd: str, logger: Optional[logging.Logger] = None):
        self.user = username
        self.passwd = passwd
        self.logger = logger or logging.getLogger('msg2jbr')
        # JabberClient gets initialized in async run() instead of __init__
        # to make sure it uses same event loop as run() instead of making its own,
        # making it possible to await JabberClient.disconnected there
        self.jabber: Optional[JabberClient] = None

    def to_be_send(self, recipient: str, message: str) -> None:
        if self.jabber is None:
            self.logger.debug(f'not running yet, message discarded: (to: {recipient}) "{message[:50]}"')
            return
        line = Line(recipient, message)
        self.jabber.send_query.put_nowait(line)

    async def run(self) -> None:
        self.jabber = JabberClient(self.user, self.passwd, self.logger.getChild('slixmpp'))
        while True:
            pending = self.jabber.send_query.qsize()
            if pending > 0:
                self.logger.debug(f'connecting to send {pending} pending messages')
                self.jabber.connect()
                try:
                    await asyncio.wait_for(self.jabber.disconnected, DISCONNECT_AFTER_DONE_DELAY * 10)
                except asyncio.TimeoutError:
                    self.logger.warning(f'sending messages takes too long, aborting to retry later')
                    await self.jabber.disconnect(wait=2.0, reason='send duration exceeded', ignore_send_queue=True)
                except (KeyboardInterrupt, asyncio.CancelledError):
                    self.logger.debug(f'interrupted while sending messages, terminating')
                    await self.jabber.disconnect(wait=2.0, reason='terminating', ignore_send_queue=True)
                    raise
                else:
                    self.logger.debug(f'done sending messages, disconnected')
            if self.jabber.fatal_error is not None:
                self.logger.warning(f'{self.jabber.fatal_error}, terminating')
                break
            pending = self.jabber.send_query.qsize()
            if pending > 0:
                self.logger.debug(f'{pending} messages left after disconnect, delaying next attempt')
                await asyncio.sleep(ON_ERROR_RETRY_DELAY)
            await asyncio.sleep(1)


class JabberClient(slixmpp.ClientXMPP):

    def __init__(self, username: str, passwd: str, logger: Optional[logging.Logger] = None) -> None:
        super().__init__(username, passwd)
        self.logger = logger or logging.getLogger('slixmppClient')
        self.send_query: asyncio.Queue = asyncio.Queue()
        self.fatal_error: Optional[str] = None
        self.add_event_handler('session_start', self.send_pending)
        self.add_event_handler('failed_all_auth', self.on_bad_auth)
        self.add_error_handlers()
        self.now_sending = False

    async def send_pending(self, _) -> None:
        self.logger.debug('got session_start event')
        if self.now_sending:
            self.logger.warning(f'session_start handler got called but now_sending={self.now_sending}, aborting')
            return
        self.now_sending = True
        try:
            self.send_presence()
            await self.get_roster()
            while True:
                line: Line = await asyncio.wait_for(self.send_query.get(), DISCONNECT_AFTER_DONE_DELAY)
                recipient = slixmpp.JID(line.recipient)
                self.send_message(mto=recipient, mbody=line.message, mtype='chat')
                self.logger.debug(f'sending message: {str(line)[:90]}')
                if self.send_query.empty():
                    self.logger.debug('put all pending messages in internal queue, waiting a bit for new ones')
        except asyncio.TimeoutError:
            self.logger.debug('disconnecting')
            await self.disconnect()
        except Exception as e:
            self.logger.exception(f'got error while sending messages: {e}')
        finally:
            self.now_sending = False

    def on_bad_auth(self, _) -> None:
        self.fatal_error = f'authentication failed for {self.boundjid.bare}'

    def add_error_handlers(self) -> None:
        error_messages = {'connection_failed': 'failed to connect',
                          'reconnect_delay': 'next connection attempt in',
                          'stream_error': 'stream error',
                          'killed': 'XML stream got aborted',
                          'message_error': 'got error message from jabber server',
                          'message': 'got message',
                          }
        for event, message in error_messages.items():
            self.add_event_handler(event, self.make_error_handler(message))

    def make_error_handler(self, message: str) -> Callable:
        def error_handler(event):
            msg = f'{message}: {event}' if event else f'{message}'
            self.logger.debug(msg)
        return error_handler



import logging

from avtdl.core.info import generate_plugins_description
from avtdl.core.loggers import silence_library_loggers


def test_config_loading(caplog):
    """
    Smoke test for plugins

    Load all plugins by generating docs,
    check it produces no warnings in log
    """
    silence_library_loggers()
    caplog.set_level(logging.WARNING)

    _ = generate_plugins_description()

    assert not caplog.records



import logging
import re
import sqlite3
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Sequence, Union

import aiohttp
import feedparser
from pydantic import ConfigDict, PositiveFloat

from avtdl.core import utils
from avtdl.core.config import Plugins
from avtdl.core.interfaces import Record, RuntimeContext
from avtdl.plugins.rss.generic_rss import GenericRSSMonitor, GenericRSSMonitorConfig, GenericRSSMonitorEntity
from avtdl.plugins.youtube.common import thumbnail_url
from avtdl.plugins.youtube.video_info import VideoInfoError, parse_video_page


@Plugins.register('rss', Plugins.kind.ASSOCIATED_RECORD)
class YoutubeFeedRecord(Record):
    """
    Youtube video or livestream parsed from a channel's RSS feed
    """
    model_config = ConfigDict(extra='allow')

    url: str
    """link to the video"""
    title: str
    """title of the video at the time of parsing"""
    published: datetime
    """published value of the feed item, usually the time when the video was uploaded or the livestream frame was set up"""
    updated: datetime
    """updated value of the feed item. If different from `published`, might indicate either a change to video title, thumbnail or description, or a change in video status, for example livestream ending"""
    thumbnail_url: Optional[str] = None
    """link to the video thumbnail"""
    author: str
    """author's name, as shown on the channel icon"""
    video_id: str
    """short string identifying the video on Youtube. Part of the video url"""
    summary: Optional[str] = None
    """video's description"""
    views: Optional[int]
    """current number of views. Is zero for upcoming and ongoing livestreams"""
    scheduled: Optional[datetime] = None
    """scheduled time for an upcoming livestream to start at, otherwise absent"""

    async def check_scheduled(self, session: Optional[aiohttp.ClientSession] = None, logger: Optional[logging.Logger] = None):
        scheduled = None
        if self.views == 0:
            logger = logger or logging.getLogger().getChild('check_scheduled')
            try:
                page = await utils.request(self.url, session, logger)
                if page is None:
                    raise VideoInfoError('failed to fetch video page')
                info = parse_video_page(page, self.url)
                scheduled = info.scheduled
            except VideoInfoError as e:
                logger.warning(f'Error while trying to check scheduled date of {self.url}, skipping')
                logger.warning(f'{e}')
            except Exception:
                logger.exception(f'Error while trying to check scheduled date of {self.url}, skipping')
        self.scheduled = scheduled

    def __str__(self):
        return self.format_record()

    def __repr__(self):
        template = '{} {:<8} [{}] {}'
        return template.format(self.format_date(self.published), self.author, self.video_id, self.title[:60])

    def get_uid(self) -> str:
        return self.video_id

    @staticmethod
    def format_date(date: Union[str, datetime]) -> str:
        if isinstance(date, str):
            date = datetime.fromisoformat(date)
        return date.strftime('%Y-%m-%d %H:%M')

    def format_record(self):
        scheduled = self.scheduled
        if scheduled:
            scheduled_time = '\nscheduled to {}'.format(self.format_date(scheduled))
        else:
            scheduled_time = ''
        template = '{}\n{}\npublished by {} at {}'
        return template.format(self.url, self.title, self.author, self.format_date(self.published)) + scheduled_time

    def as_dict(self, additional_fields):
        record_dict = {}
        record_dict.update(self.__dict__)
        record_dict.update(additional_fields)
        return record_dict

    def discord_embed(self) -> dict:
        embed: Dict[str, Any] = {
            'title': self.title,
            # 'description': ,
            'url': self.url,
            'color': None,
            'author': {'name': self.author},
            'timestamp': self.format_date(self.published),
            'image': {'url': thumbnail_url(self.video_id)}
        }
        if self.scheduled is not None:
            scheduled = self.scheduled.strftime('%Y-%m-%d %H:%M')
            embed['fields'] = [{'name': 'Scheduled:', 'value': scheduled, 'inline': True}]
        return embed


@Plugins.register('rss', Plugins.kind.ACTOR_ENTITY)
class FeedMonitorEntity(GenericRSSMonitorEntity):
    update_interval: PositiveFloat = 900
    """How often the feed should be updated, in seconds"""
    track_reschedule: bool = False
    """keep track of scheduled time of upcoming streams, emit record again if it is changed to an earlier date"""


@Plugins.register('rss', Plugins.kind.ACTOR_CONFIG)
class FeedMonitorConfig(GenericRSSMonitorConfig):
    pass


@Plugins.register('rss', Plugins.kind.ACTOR)
class FeedMonitor(GenericRSSMonitor):
    """
    Youtube channel RSS feed monitor

    Monitors channel for new uploads and livestreams using RSS feed
    generated by Youtube. Requires old channel id format. In order to obtain
    RSS feed url for a given channel, use "View Source" on a channel page
    and search for "rss".

    Example of a supported url:

    - `https://www.youtube.com/feeds/videos.xml?channel_id=UCK0V3b23uJyU4N8eR_BR0QA`

    RSS feed is smaller and faster to parse compared to an HTML channel page,
    but by design only shows updates of a single channel and doesn't support
    authentication and therefore unable to show member-only streams.

    Scheduled date for upcoming streams is not present in the feed itself, so it
    is obtained by fetching and parsing video page the first time it appears in
    the feed. It then gets updated until stream goes live, unless `track_reschedule`
    option is disabled.

    No matter how often the url gets fetched, content of the feed only gets
    changed once every 15 minutes, so setting `update_interval` lower than that
    value is not recommended. This monitor will attempt to calculate time of the
    next update from HTTP headers and schedule next request right after it. Use
    `adjust_update_interval` to disable this behavior.
    """

    def __init__(self, conf: FeedMonitorConfig, entities: Sequence[FeedMonitorEntity], ctx: RuntimeContext):
        try:
            Migration(conf.db_path)
        except Exception as e:
            raise Exception('migration failed') from e
        super().__init__(conf, entities, ctx)

    async def get_records(self, entity: FeedMonitorEntity, session: aiohttp.ClientSession) -> Sequence[YoutubeFeedRecord]:
        records = await super().get_records(entity, session)
        return records

    async def get_new_records(self, entity: FeedMonitorEntity, session: aiohttp.ClientSession) -> Sequence[YoutubeFeedRecord]:
        records = await self.get_records(entity, session)
        rescheduled_records: List[YoutubeFeedRecord] = []

        # record.check_scheduled() involves loading video page, so it should only be done when necessarily
        # and here seems to be the only good place for it, since network request requires "session" object
        for record in records:
            previous = self.load_record(record, entity)
            if previous is None:
                await record.check_scheduled(session, self.logger)
                continue
            if not entity.track_reschedule:
                continue
            if not isinstance(previous, YoutubeFeedRecord):
                self.logger.warning(f'[{entity.name}] previous version of record "{record!r}" is not YoutubeFeedRecord: {previous.model_dump()}')
                continue
            if previous is not None and previous.scheduled is not None:
                self.logger.debug(f'{record.video_id=} has last {previous.scheduled=}, updating')
                await record.check_scheduled(session, self.logger)
                if record.scheduled is None:
                    continue
                if record.scheduled < previous.scheduled:
                    # treat rescheduled records as new if scheduled time is earlier than before
                    # to allow action run on time, though it will run second time later
                    msg = 'In feed {} record [{}] {} rescheduled back from {} to {}'
                    msg = msg.format(entity.name, record.video_id, record.title, previous.scheduled, record.scheduled)
                    self.logger.warning(msg)
                    rescheduled_records.append(record)

        new_records = self.filter_new_records(records, entity)
        rescheduled_records.extend(new_records)
        return rescheduled_records

    def record_got_updated(self, record: YoutubeFeedRecord, entity: FeedMonitorEntity) -> bool:
        excluded_fields = {'views'}
        return self.db.record_has_changed(record, entity.name, excluded_fields)

    def _parse_entry(self, entry: feedparser.FeedParserDict) -> YoutubeFeedRecord:
        parsed: Dict[str, Any] = {}
        parsed['url'] = entry['link']
        parsed['title'] = entry['title']
        parsed['updated'] = utils.make_datetime(entry['updated_parsed'])

        parsed['published'] = utils.make_datetime(entry['published_parsed'])
        parsed['author'] = entry.get('author')
        parsed['summary'] = entry.get('summary')
        video_id = entry.get('yt_videoid')
        if video_id is not None:
            parsed['video_id'] = video_id
            parsed['thumbnail_url'] = thumbnail_url(video_id)
        else:
            parsed['video_id'] = None
            parsed['thumbnail_url'] = None
            self.logger.warning(f'[{entry.name}] got entry without video_id: {entry}')
        """link to the video thumbnail"""
        try:
            views = int(entry['media_statistics']['views'])
        except (ValueError, KeyError, TypeError):
            views = None
        parsed['views'] = views
        record = YoutubeFeedRecord(**parsed)
        return record


class Migration:
    """Convert db produced by RSS2JBR in format used by RecordDB"""

    name: str = 'rss2jbr'
    table_name = 'records'
    target_table_structure = 'parsed_at datetime, feed_name text, uid text, hashsum text, class_name text, as_json text, PRIMARY KEY(uid, hashsum)'
    target_row_structure = ':parsed_at, :feed_name, :uid, :hashsum, :class_name, :as_json'

    migrated_table_schema = '''CREATE TABLE "records" ( "parsed_at" datetime, "feed_name" text, "author" text, "video_id" text, "link" text, "title" text, "summary" text, "published" datetime, "updated" datetime, "scheduled" datetime DEFAULT NULL, "views" intefer, PRIMARY KEY("video_id","updated"))'''

    def __init__(self, db_path: Union[str, Path]):
        self.logger = logging.getLogger('migration').getChild(self.name)
        dt = datetime.now(tz=timezone.utc).strftime('%Y_%m_%d_%H_%M_%S_%f')
        self.backup_table_name = f'migration_{self.table_name}_{dt}'
        if db_path == ':memory:':
            return
        if not Path(db_path).exists():
            return
        try:
            self.db = sqlite3.connect(db_path, isolation_level=None)
            self.db.row_factory = sqlite3.Row
        except sqlite3.OperationalError as e:
            self.logger.error(f'error opening sqlite database at path "{db_path}": {e}')
            raise
        else:
            self.logger.debug(f'successfully connected to sqlite database at "{db_path}"')

        with self.db:
            self.db.execute('BEGIN TRANSACTION')

            if not self.is_valid_target(self.table_name):
                self.logger.info(f'no migration is applicable for table "{self.table_name}" at "{db_path}"')
                return

            self.logger.info(f'running migration for table "{self.table_name}" at "{db_path}"')
            self.migrate()
            self.logger.info(f'migration for table "{self.table_name}" in file "{db_path}" is completed. '
                             f'Original data are preserved at table "{self.backup_table_name}"')

    def migrate(self):
        self.rename_table(self.table_name, self.backup_table_name)
        self.init_target_table(self.table_name)
        max_bad_records = 5
        for row in self.fetch_all_rows(self.backup_table_name):
            try:
                converted_row = self.transform(row)
            except Exception as e:
                self.logger.warning(f'failed to transform a row: {e}. Skipping failed record. Raw row: "{dict(row)}"')
                max_bad_records -= 1
                if max_bad_records <= 0:
                    raise Exception('exceeded limit for bad records, aborting migration') from e
                continue
            self.store(converted_row)

    def get_table_schema(self, table_name) -> Optional[str]:
        sql = 'SELECT sql FROM sqlite_schema WHERE name = :table_name'
        keys = {'table_name': table_name}
        cursor = self.db.execute(sql, keys)
        row = cursor.fetchone()
        if row is None:
            return None
        return row['sql']
    
    @staticmethod
    def _normalize_schema(schema: str) -> str:
        schema = schema.lower()
        schema = re.sub('[\s"\'`]', '', schema) 
        schema = schema.replace('intefer', 'integer')
        return schema

    def is_valid_target(self, table_name: str) -> bool:
        """returns True if this migration can be applied to given database"""
        existing_schema = self.get_table_schema(table_name) or ''
        return self._normalize_schema(self.migrated_table_schema) == self._normalize_schema(existing_schema)

    def rename_table(self, old: str, new: str):
        sql = f'ALTER TABLE {old} RENAME TO {new}'
        self.db.execute(sql)

    def init_target_table(self, table_name: str):
        self.db.execute('CREATE TABLE IF NOT EXISTS {} ({})'.format(table_name, self.target_table_structure))

    def fetch_all_rows(self, table_name: str) -> Iterator[sqlite3.Row]:
        sql = f'SELECT * FROM {table_name}'
        cursor = self.db.execute(sql)
        while True:
            row = cursor.fetchone()
            if row is None:
                break
            yield row

    @staticmethod
    def load_record(row: sqlite3.Row) -> YoutubeFeedRecord:
        row = dict(row)
        for field in ['parsed_at', 'feed_name']:
            row.pop(field)
        row['url'] = row.pop('link')
        for field in ['published', 'updated', 'scheduled']:
            value = row[field]
            # handle db records with old format
            if isinstance(value, str):
                row[field] = datetime.fromisoformat(value)
        return YoutubeFeedRecord(**row)

    def transform(self, row: sqlite3.Row) -> Optional[Dict[str, Any]]:
        """transform row in existing format into a row in target format"""
        record = self.load_record(row)

        transformed_row = {
            'parsed_at': row['parsed_at'],
            'feed_name': row['feed_name'],
            'uid': '{}:{}'.format(row['feed_name'], row['video_id']),
            'hashsum': record.hash(),
            'class_name': record.__class__.__name__,
            'as_json': record.as_json(),
        }
        return transformed_row

    def store(self, rows: Union[Dict[str, Any], List[Dict[str, Any]]]) -> None:
        sql = "INSERT OR IGNORE INTO {} VALUES({})".format(self.table_name, self.target_row_structure)
        if not isinstance(rows, list):
            rows = [rows]
        self.db.executemany(sql, rows)



import asyncio
import datetime
import json
import logging
from textwrap import shorten
from typing import Dict, List, Optional, Sequence, Tuple, Union

import aiohttp
import multidict

from avtdl.core.interfaces import Action, ActionEntity, ActorConfig, Record, RuntimeContext
from avtdl.core.plugins import Plugins
from avtdl.core.utils import RateLimit, request_raw

EMBEDS_PER_MESSAGE = 10
EMBED_TITLE_MAX_LENGTH = 256
EMBED_DESCRIPTION_MAX_LENGTH = 4096


class DiscordRateLimit(RateLimit):

    def _submit_headers(self, headers: Union[Dict[str, str], multidict.CIMultiDictProxy[str]], logger: logging.Logger):
        try:
            self.limit_total = int(headers.get('X-RateLimit-Limit', -1))
            self.limit_remaining = int(headers.get('X-RateLimit-Remaining', -1))
            self.reset_at = int(headers.get('X-RateLimit-Reset', -1))
        except ValueError:
            logger.warning(f'[{self.name}] error parsing rate limit headers: "{headers}"')
        else:
            logger.debug(f'[{self.name}] rate limit {self.limit_remaining}/{self.limit_total}, resets after {datetime.timedelta(seconds=self.reset_after)}')

    @staticmethod
    def get_bucket(headers: multidict.CIMultiDictProxy[str]) -> Optional[str]:
        return headers.get('X-RateLimit-Bucket')


class NoRateLimit(DiscordRateLimit):

    def _submit_headers(self, headers: Union[Dict[str, str], multidict.CIMultiDictProxy[str]], logger: logging.Logger):
        pass


class MessageFormatter:

    @classmethod
    def format(cls, records: List[Record]) -> Tuple[dict, List[Record]]:
        '''take records and format them in Discord webhook payload as embeds
        after the limit on embeds is reached, the rest of the records are returned back'''
        embeds: List[dict] = []
        excess_records = []
        for i, record in enumerate(records):
            record_embeds = cls.make_embeds(record)
            if len(embeds) + len(record_embeds) > EMBEDS_PER_MESSAGE:
                excess_records = records[i:]
                break
            else:
                embeds.extend(record_embeds)
        message = cls.make_message(embeds)
        return message, excess_records

    @classmethod
    def make_message(cls, embeds: List[dict]) -> dict:
        return {
            "content": None,
            "embeds": embeds
        }

    @classmethod
    def make_embeds(cls, record: Record) -> List[dict]:
        formatter = getattr(record, 'discord_embed', None)
        if formatter is not None and callable(formatter):
            embeds = formatter()
            if not isinstance(embeds, list):
                embeds = [embeds]
            return embeds
        return [cls.plaintext_embed(record)]

    @classmethod
    def plaintext_embed(cls, record: Record) -> dict:
        text = str(record)
        if text.find('\n') > -1:
            title, description = text.split('\n', 1)
        else:
            title, description = '', text
        title = shorten(title, EMBED_TITLE_MAX_LENGTH)
        description = shorten(description, EMBED_DESCRIPTION_MAX_LENGTH)
        return {'title': title, 'description': description}

    @classmethod
    def check_limits(cls, message: dict) -> bool:
        # doesn't count field.name and field.value number and size in hope it will not change outcome

        class Limits:
            TOTAL = 6000
            AUTHOR_NAME = 256
            TITLE = 256
            DESCRIPTION = 4096
            FOOTER_TEXT = 2048

        total_length = 0
        embeds = message.get('embeds', [])
        for embed in embeds:
            author_name = len(embed.get('author', {}).get('name', '') or '')
            title = len(embed.get('title') or '')
            description = len(embed.get('description') or '')
            footer_text = len((embed.get('footer') or {}).get('text') or '')
            if author_name > Limits.AUTHOR_NAME:
                return False
            if title > Limits.TITLE:
                return False
            if description > Limits.DESCRIPTION:
                return False
            if footer_text > Limits.FOOTER_TEXT:
                return False
            total_length += author_name + title + description + footer_text

        if total_length > Limits.TOTAL:
            return False
        return True


@Plugins.register('discord.hook', Plugins.kind.ACTOR_CONFIG)
class DiscordHookConfig(ActorConfig):
    pass


@Plugins.register('discord.hook', Plugins.kind.ACTOR_ENTITY)
class DiscordHookEntity(ActionEntity):
    url: str
    """webhook url"""


@Plugins.register('discord.hook', Plugins.kind.ACTOR)
class DiscordHook(Action):
    """
    Send record to Discord using webhook

    To generate webhook url follow instructions in "Making a Webhook" section of
    <https://support.discord.com/hc/en-us/articles/228383668-Intro-to-Webhooks>

    Some record types support rich formatting when sent to Discord, such as
    showing the author's avatar and links to attached images. Youtube videos will
    show thumbnail, however embedding video itself is not supported.

    Records coming within six seconds one after another will be batched together into a single message.
    When too many records are received at once, they will be sent with delays to conform to Discord
    rate limits. Records deemed to be too long to fit in a Discord message
    [length limits](https://discord.com/developers/docs/resources/channel#create-message-jsonform-params)
    will be dropped with a warning.
    """

    def __init__(self, conf: DiscordHookConfig, entities: Sequence[DiscordHookEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)
        self.queues: Dict[str, asyncio.Queue] = {entity.name: asyncio.Queue() for entity in entities}
        self.buckets: Dict[Optional[str], DiscordRateLimit] = {}
        self.buckets[None] = NoRateLimit('fallback bucket')

    def handle(self, entity: DiscordHookEntity, record: Record):
        if not entity.name in self.queues:
            return
        self.logger.debug(f'[{entity.name}] adding record to send query: {record!r}')
        self.queues[entity.name].put_nowait(record)

    async def run(self):
        async with aiohttp.ClientSession() as session:
            for entity in self.entities.values():
                _ = self.controller.create_task(self.run_for(entity, session), name=f'{self.conf.name}:{entity.name}')
            await asyncio.Future()

    async def run_for(self, entity: DiscordHookEntity, session: aiohttp.ClientSession):
        send_queue = self.queues[entity.name]

        bucket: Optional[str] = None
        until_next_try = 0
        to_be_sent: List[Record] = []

        while True:
            await asyncio.sleep(until_next_try)
            to_be_sent = await self.wait_for_records(send_queue, pending=to_be_sent)
            message, pending_records = self._prepare_message(to_be_sent, entity)
            if message is None:
                to_be_sent = pending_records
                continue

            async with self.buckets[bucket] as _:
                headers = {'Content-Type': 'application/json'}
                try:
                    response = await request_raw(entity.url, session, self.logger,
                                                 method='POST',
                                                 data=json.dumps(message),
                                                 headers=headers,
                                                 raise_errors=True,
                                                 raise_for_status=False)
                    assert response is not None, 'request_json() returned None despite raise_errors=True'
                except (OSError, asyncio.TimeoutError, aiohttp.ClientConnectionError) as e:
                    self.logger.warning(f'[{entity.name}] error while sending message with Discord webhook: {e or type(e)}, saving for the next try')
                    until_next_try = 60
                    continue  # implicitly saving messages in to_be_sent until the next try
                except Exception as e:
                    self.logger.exception(f'[{entity.name}] error while sending message with Discord webhook: {e or type(e)}')
                    self.logger.debug(f'raw message text:\n{message}')
                    to_be_sent = pending_records # discard unsent messages as they might have been the cause of error
                    until_next_try = 60
                    continue

            bucket = DiscordRateLimit.get_bucket(response.headers)
            if not bucket in self.buckets:
                self.buckets[bucket] = DiscordRateLimit(f'{bucket}', self.logger)
            self.buckets[bucket].submit_headers(response.headers, self.logger)

            if self._has_fatal_error(response, entity, message):
                break

            # if message got send successfully discard records except these that didn't fit into message
            if pending_records:
                self.logger.debug(f'carrying {len(pending_records)} pending records to the next loop')
            to_be_sent = pending_records

    @staticmethod
    async def wait_for_records(queue: asyncio.Queue, pending: List[Record]) -> List[Record]:
        to_be_sent = pending
        while len(to_be_sent) < EMBEDS_PER_MESSAGE:
            try:
                record = await asyncio.wait_for(queue.get(), 60 / EMBEDS_PER_MESSAGE)
                to_be_sent.append(record)
            except asyncio.TimeoutError:
                if to_be_sent:
                    break
        return to_be_sent

    def _prepare_message(self, to_be_sent: List[Record], entity: DiscordHookEntity) -> Tuple[Optional[dict], List[Record]]:
        try:
            message, pending_records = MessageFormatter.format(to_be_sent)
            if pending_records:
                self.logger.debug(f'[{entity.name}] out of {len(to_be_sent)} records {len(pending_records)} are left pending')
        except Exception as e:
            self.logger.exception(f'error happened while formatting message: {e}\nRaw records list: "{to_be_sent}"')
            return None, []
        try:
            limits_ok = MessageFormatter.check_limits(message)
        except Exception as e:
            self.logger.exception(f'error checking content length limits for message: {e}\nRaw message: "{message}"')
            return None, pending_records
        if not limits_ok:
            self.logger.warning(f'[{entity.name}] prepared message exceeded Discord length limits. Records will be discarded')
            self.logger.debug(f'[{entity.name}] message content:\n{message}')
            return None, pending_records

        return message, pending_records

    def _has_fatal_error(self, response: aiohttp.ClientResponse, entity: DiscordHookEntity, message: dict) -> bool:
        try:
            response.raise_for_status()
            return False
        except aiohttp.ClientResponseError as e:
            self.logger.warning(f'[{entity.name}] failed to send message: got {response.status} ({response.reason}) from {entity.url}')
            if e.status == 400:
                self.logger.warning(f'[{entity.name}] message got rejected with {response.status} ({response.reason}), dropping it')
                self.logger.debug(f'[{entity.name}] request headers: {response.request_info.headers}')
                self.logger.debug(f'[{entity.name}] response headers: {response.headers}')
                self.logger.debug(f'[{entity.name}] raw request body: {json.dumps(message)}')
                return False
            elif e.status in [404, 401]:
                self.logger.warning(f'[{entity.name}] got {e.status} from webhook, interrupting operations. Check if webhook url is still valid: {entity.url}')
                return True
            else:
                return False



import asyncio
import json
import logging
import pathlib
from typing import Dict, List, Optional

import dateutil.zoneinfo
from aiohttp import web
from pydantic import BaseModel

from avtdl.core import info
from avtdl.core.chain import Chain
from avtdl.core.config import ConfigParser, ConfigurationError, SettingsSection
from avtdl.core.info import get_known_plugins, get_plugin_type, render_markdown
from avtdl.core.interfaces import Actor, Record, RuntimeContext, TerminatedAction
from avtdl.core.plugins import Plugins
from avtdl.core.utils import strip_text, write_file
from avtdl.core.yaml import merge_data, yaml_dump


def serialize_config(settings: SettingsSection,
                     actors: Dict[str, Actor],
                     chains: Dict[str, Chain]) -> str:
    config = ConfigParser.serialize(settings, actors, chains)
    conf = config.model_dump_json()
    return conf


def json_dumps(obj):
    def default(o):
        if isinstance(o, pathlib.Path):
            return str(o)
        raise TypeError(f'Object of type {o.__class__.__name__} is not JSON serializable')

    return json.dumps(obj, default=default)


def get_actor_description(name: str) -> str:
    plugin, config, entity = Plugins.get_actor_factories(name)
    md_description = info.render_doc(plugin)
    html_description = info.render_markdown(md_description)
    return html_description


def get_schema(model: BaseModel) -> dict:
    schema = model.model_json_schema(mode='serialization')
    render_descriptions(schema)
    return schema


def render_descriptions(schema) -> None:
    if isinstance(schema, dict):
        if 'description' in schema:
            text = schema['description']
            if isinstance(text, str) and len(text) > 0:
                end = '' if text[-1] in '.?!,' else '.'
                text = text[0].upper() + text[1:] + end
            schema['description'] = info.render_markdown(text)
        else:
            for subschema in schema.values():
                render_descriptions(subschema)
    elif isinstance(schema, list) or isinstance(schema, tuple):
        for subschema in schema:
            render_descriptions(subschema)
    else:
        return


def get_conf_schema(actor_name: str) -> dict:
    plugin, config, entity = Plugins.get_actor_factories(actor_name)
    return get_schema(config)


def get_entity_schema(actor_name: str) -> dict:
    plugin, config, entity = Plugins.get_actor_factories(actor_name)
    return get_schema(entity)


def record_preview(record: Record, representation: str = 'text') -> str:
    if representation == 'text':
        return str(record).replace('\n', '<br>\n')
    elif representation == 'json':
        return record.as_json(indent=4)
    elif representation == 'short':
        return repr(record)
    else:
        return str(record)


class ActorModel(BaseModel):
    type: Optional[str]
    description: str
    config_schema: dict
    entity_schema: dict


class WebUI:
    WEBROOT: pathlib.Path = pathlib.Path(__file__).parent.parent.resolve() / 'ui'
    RESTART_DELAY: int = 3

    def __init__(self, config_path: pathlib.Path, config, ctx: RuntimeContext, settings: SettingsSection, actors: Dict[str, Actor], chains: Dict[str, Chain]):
        self.logger = logging.getLogger('webui')
        self.host = settings.host
        self.port = settings.port
        self.config_path = config_path
        self.config_base = config
        self.ctx = ctx
        self.settings = settings
        self.actors = actors
        self.chains = chains
        self.restart_pending = False
        self._actors_models = self.generate_actors_models()
        self.routes: List[web.AbstractRouteDef] = []

        self.routes.append(web.get('/favicon.ico', self.favicon))
        self.routes.append(web.get('/chains', self.show_chains))
        self.routes.append(web.get('/actors', self.actors_models))
        self.routes.append(web.get('/settings', self.settings_schema))
        self.routes.append(web.get('/config', self.show_config))
        self.routes.append(web.post('/config', self.store_config))
        self.routes.append(web.get('/timezones', self.timezones))
        self.routes.append(web.get('/motd', self.motd))
        self.routes.append(web.get('/history', self.history))

        self.routes.append(web.get('/ui/info/info.html', self.info_webui))

        self.routes.append(web.get('/', self.index))
        self.routes.append(web.static('/ui', self.WEBROOT))

    async def favicon(self, request: web.Request) -> web.Response:
        return web.Response()

    async def index(self, request):
        raise web.HTTPFound('/ui/conf/config.html')

    async def timezones(self, request: web.Request) -> web.Response:
        zones = list(dateutil.zoneinfo.get_zonefile_instance().zones.keys())
        return web.json_response(zones)

    async def show_chains(self, request: web.Request) -> web.Response:
        data = {name: chain.conf.model_dump() for name, chain in self.chains.items()}
        return web.json_response(data)

    @staticmethod
    def generate_actors_models() -> Dict[str, dict]:
        data: Dict[str, dict] = {}

        for name in get_known_plugins():
            data[name] = ActorModel(
                type=get_plugin_type(name),
                description=get_actor_description(name),
                config_schema=get_conf_schema(name),
                entity_schema=get_entity_schema(name)
            ).model_dump()
        return data

    async def actors_models(self, request: web.Request) -> web.Response:
        return web.json_response(self._actors_models, dumps=json_dumps)

    async def settings_schema(self, request: web.Request) -> web.Response:
        schema = self.settings.model_json_schema(mode='serialization')
        render_descriptions(schema)
        return web.json_response(schema, dumps=json_dumps)

    async def show_config(self, request: web.Request):
        data = serialize_config(self.settings, self.actors, self.chains)
        return web.Response(text=data)

    async def store_config(self, request: web.Request) -> web.Response:
        mode = request.query.get('mode', 'check')
        if not mode in ['check', 'store', 'reload']:
            raise web.HTTPBadRequest(text=f'unexpected "mode" parameter {mode}')
        try:
            conf = await request.json()
            _ = ConfigParser.validate(conf)
            updated_config = merge_data(self.config_base, conf)
            raw_yaml = yaml_dump(updated_config)
            if mode == 'check':
                return web.Response(text='Config has been validated successfully')

            self.config_base = updated_config

            try:
                write_file(self.config_path, raw_yaml, backups=10)
            except Exception as e:
                raise web.HTTPInternalServerError(text=f'failed to store config in "{self.config_path}": {e or type(e)}')
            if mode == 'reload':
                self.restart_pending = True
                self.ctx.controller.terminate_after(self.RESTART_DELAY, TerminatedAction.RESTART)
                return web.Response(text=f'Updated config successfully stored in "{self.config_path}". Restarting in a few seconds.')
            else:
                text = f'Updated config successfully stored in "{self.config_path}". It will be used after next restart.'
                return web.Response(text=text)
        except ConfigurationError as e:
            if e.__cause__ is None:
                raise web.HTTPBadRequest(text=f'Malformed configuration error {type(e)}: {e}')
            data = e.__cause__.errors()
            for error in data:
                if 'url' in error:
                    error.pop('url')
                if 'ctx' in error:
                    error.pop('ctx')
                if 'msg' in error:
                    error['msg'] = strip_text(error['msg'], 'Value error, ')

            return web.json_response(data=data, dumps=json_dumps, status=422, reason='Bad config')
        except web.HTTPError:
            raise
        except Exception as e:
            raise web.HTTPBadRequest(text=f'{type(e)}: {e}')
        raise web.HTTPFound(location=request.path, reason='Config OK')

    async def motd(self, request: web.Request) -> web.Response:
        if self.restart_pending:
            raise web.HTTPServiceUnavailable(headers={'Retry-After': str(self.RESTART_DELAY)})
        motd = f'''
Server is up and running, working directory is "{pathlib.Path('.').resolve()}".
Configuration contains {len(self.actors)} actors and {len(self.chains)} chains, loaded from "{self.config_path.resolve()}".
'''
        data = {'motd': motd}
        return web.json_response(data, dumps=json_dumps)

    async def info_webui(self, request: web.Request) -> web.Response:

        template_path = self.WEBROOT / 'info/info.html'
        template = template_path.read_text(encoding='utf8')

        document_path = self.WEBROOT / 'info/info.md'
        document = document_path.read_text(encoding='utf8')
        body = render_markdown(document)

        html = template.replace('{{body}}', body)
        return web.Response(text=html, content_type='text/html')

    async def history(self, request: web.Request) -> web.Response:
        actor = request.query.get('actor')
        entity = request.query.get('entity')
        chain = request.query.get('chain', '')
        representation = request.query.get('repr', 'text')
        if actor is None or entity is None:
            raise web.HTTPBadRequest(text=f'not enough arguments. Got actor="{actor}", entity="{entity}", chain="{chain}"')
        incoming = self.ctx.bus.get_history(actor, entity, chain, 'in')
        outgoing = self.ctx.bus.get_history(actor, entity, chain, 'out')
        data_structure = [
            (f'Incoming records (most recent)', incoming),
            (f'Outgoing records (most recent)', outgoing)
        ]
        data = {}

        for title, content in data_structure:
            records = [[record.origin, record.chain, record_preview(record, representation)] for record in content]
            data[title] = records
        return web.json_response(data, dumps=json_dumps)


async def run_app(webui: WebUI):
    app = web.Application()
    app.add_routes(webui.routes)
    runner = web.AppRunner(app)
    await runner.setup()
    site = web.TCPSite(runner, webui.host, webui.port)
    webui.logger.debug('starting server...')
    try:
        await site.start()
    except Exception as e:
        webui.logger.exception(f'failed to start server: {e}')
        return
    webui.logger.info(f'server is running on http://{webui.host}:{webui.port}')
    try:
        await asyncio.Future()
    except (KeyboardInterrupt, asyncio.CancelledError):
        webui.logger.info('stopping server...')
        await runner.cleanup()
        webui.logger.debug('server stopped')


async def run(config_path: pathlib.Path, config, ctx: RuntimeContext, settings: SettingsSection, actors: Dict[str, Actor], chains: Dict[str, Chain]):
    webui = WebUI(config_path, config, ctx, settings, actors, chains)
    await run_app(webui)



import datetime
from typing import List

import pytest

from avtdl.core.interfaces import Event, EventType, RuntimeContext, TextRecord
from avtdl.plugins.filters.filters import EmptyFilterConfig, EmptyFilterEntity, EventCauseFilter, FormatEventFilter, FormatEventFilterEntity, MatchFilter, MatchFilterEntity
from avtdl.plugins.twitch.twitch import TwitchRecord


@pytest.fixture()
def empty_config():
    return EmptyFilterConfig(name='test')


@pytest.fixture()
def empty_entity():
    return EmptyFilterEntity(name='test')


@pytest.fixture()
def text_record():
    return TextRecord(text='test')


@pytest.fixture()
def event_record(text_record):
    return Event(event_type=EventType.error, text='test event', record=text_record)


class TestEventCause:

    @staticmethod
    @pytest.fixture()
    def cause_filter(empty_config, empty_entity):
        ctx = RuntimeContext.create()
        return EventCauseFilter(config=empty_config, entities=[empty_entity], ctx=ctx)

    @staticmethod
    def test_match(cause_filter, empty_entity, event_record):
        result = cause_filter.match(empty_entity, event_record)
        assert result == event_record.record

    @staticmethod
    def test_passthrough(cause_filter, empty_entity, text_record):
        assert cause_filter.match(empty_entity, text_record) == text_record


class TestFormatEvent:

    @staticmethod
    def prepare_filter():
        empty_config = EmptyFilterConfig(name='test')
        entity = FormatEventFilterEntity(name='test', type_template='error', text_template='Event: *{text}*')
        ctx = RuntimeContext.create()
        return FormatEventFilter(config=empty_config, entities=[entity], ctx=ctx), entity

    def test_match(self, text_record):
        format_event_filter, entity = self.prepare_filter()
        result = format_event_filter.match(entity, text_record)
        assert isinstance(result, Event)
        assert result.record == text_record
        assert result.event_type == 'error'
        assert result.text == 'Event: *test*'


class TestMatchFilter:

    @staticmethod
    def prepare_filter(patterns: List[str], fields: List[str]):
        config = EmptyFilterConfig(name='test')
        entity = MatchFilterEntity(name='test', patterns=patterns, fields=fields)
        ctx = RuntimeContext.create()
        return MatchFilter(config, [entity], ctx), entity

    def test_empty_field_matches(self):
        record = TwitchRecord(
            url='https://twitch.tv/test',
            username='test',
            title='test',
            start=datetime.datetime.fromtimestamp(0),
            avatar_url='https://example.com/'
        )
        match_filter, entity = self.prepare_filter([''], ['game'])

        result = match_filter.match(entity, record)
        assert result == record



import logging
from collections import Counter, defaultdict
from typing import Callable, List, OrderedDict

from pydantic import RootModel, field_validator

from avtdl.core.interfaces import Record, RuntimeContext


class CardSection(RootModel):
    """Single element of Chain"""
    root: OrderedDict[str, List[str]]

    def get_item(self) -> tuple[str, list[str]]:
        return self.root.copy().popitem()

    @field_validator('root')
    @classmethod
    def check_card(cls, root):
        if len(root) != 1:
            raise ValueError(f'card must list exactly 1 actor, got {len(root)}')
        actor_name = list(root.keys())[0]
        entities_names = root[actor_name]
        if len(entities_names) < 1:
            raise ValueError(f'{actor_name}: should list at least one entity name')
        return root


class ChainConfigSection(RootModel):

    root: List[CardSection]

    def __iter__(self):
        for card in self.root:
            try:
                yield card.get_item()
            except KeyError:
                continue

    def __len__(self):
        return self.root.__len__()

    def __getitem__(self, item):
        value = self.root.__getitem__(item)
        if isinstance(value, list):
            return [card.get_item() for card in value if card]
        return value.get_item()

    @field_validator('root')
    @classmethod
    def check_chains(cls, values):
        if len(values) < 2:
            raise ValueError('chain must contain at least 2 elements')
        return values


class Chain:

    def __init__(self, name: str, actors: ChainConfigSection, ctx: RuntimeContext):
        self.name = name
        self.bus = ctx.bus
        self.logger = logging.getLogger('chain')
        self.conf = actors

        if len(actors) < 2:
            self.logger.warning(f'chain {name}: need at least two actors to create a chain')
            return

        self.check_for_duplicated_entities(name, actors)

        producer_name, producer = actors[0]
        for consumer_name, consumer in actors[1:]:
            for producer_entity in producer:
                for consumer_entity in consumer:
                    producer_topic = self.bus.outgoing_topic_for(producer_name, producer_entity, self.name)
                    consumer_topic = self.bus.incoming_topic_for(consumer_name, consumer_entity, self.name)
                    handler = self.get_handler(consumer_topic)
                    self.bus.sub(producer_topic, handler)
            producer_name, producer = consumer_name, consumer

    def check_for_duplicated_entities(self, chain_name, actors: ChainConfigSection) -> None:
        flattened_actors = defaultdict(list)
        for name, entities in actors:
            flattened_actors[name].extend(entities)
        counted_actors = {}
        for name, entities in flattened_actors.items():
            counted_actors[name] = Counter(entities)
        for name, counter in counted_actors.items():
            for entity_name, times in counter.most_common():
                if times > 1:
                    msg = f'Chain {chain_name}: {name}: {entity_name} is used multiple times. It might lead to infinite recursion (WILL lead for filters), causing crash or triggering OOM killer. Remove duplicates from the chain or make each of them a separate entity with different name'
                    raise ValueError(msg)

    def get_handler(self, topic) -> Callable[[str, Record], None]:
        # noinspection PyMethodParameters
        class Handler:
            def __init__(this):
                this.logger = self.logger.getChild('handler')

            def __call__(this, producer_topic: str, record: Record):
                this.logger.debug(f'Chain({self.name}): from {producer_topic} to {topic} forwarding record "{record!r}"')
                self.bus.pub(topic, record)

            def __repr__(this):
                return f'Chain({self.name}).handler({topic})'

        return Handler()

    def __repr__(self):
        actors_names = [name for name, _ in self.conf]
        return f'Chain("{self.name}", {actors_names})'



import asyncio
import datetime
import json
import logging
import signal
from abc import ABC, abstractmethod
from collections import defaultdict, deque
from enum import Enum
from hashlib import sha1
from textwrap import shorten
from typing import Any, Callable, Coroutine, Deque, Dict, List, Literal, Optional, Sequence, Tuple, Union

import dateutil.tz
from pydantic import BaseModel, ConfigDict, Field, SerializeAsAny, field_serializer, field_validator

MAX_REPR_LEN = 60


class Record(BaseModel):
    '''Data entry, passed around from Monitors to Actions through Filters'''

    model_config = ConfigDict(use_attribute_docstrings=True)

    origin: Optional[str] = Field(default=None, exclude=True)
    """semicolon-separated names of actor and entity record originated from"""
    chain: str = Field(default='', exclude=True)
    """name of the Chain this record is going through.
    Empty string means it was just produced and should go to every subscriber"""
    created_at: datetime.datetime = Field(default_factory=datetime.datetime.now, exclude=True)
    """record creation timestamp"""
    @abstractmethod
    def __str__(self) -> str:
        '''Text representation of the record to be sent in a message, written to a file etc.'''

    @abstractmethod
    def __repr__(self) -> str:
        '''Short text representation of the record to be printed in logs'''

    def __eq__(self, other) -> bool:
        if not isinstance(other, Record):
            return NotImplemented
        return self.model_dump() == other.model_dump()

    def get_uid(self) -> str:
        '''A string that is the same for different versions of the same record'''
        return self.hash()

    def as_timezone(self, timezone: Optional[datetime.timezone] = None) -> 'Record':
        fields = dict(self)
        for k, v in fields.items():
            if isinstance(v, Record):
                fields[k] = v.as_timezone(timezone)
            if isinstance(v, datetime.datetime):
                fields[k] = v.astimezone(timezone)
        record_copy = self.model_validate(fields)
        return record_copy

    def as_json(self, indent: Union[int, str, None] = None) -> str:
        return json.dumps(self.model_dump(), sort_keys=True, ensure_ascii=False, default=str, indent=indent)

    def hash(self) -> str:
        record_hash = sha1(self.as_json().encode())
        return record_hash.hexdigest()


class TextRecord(Record):
    """
    Simplest record, containing only a single text field
    """

    text: str
    """content of the record"""

    def __str__(self) -> str:
        return self.text

    def __repr__(self) -> str:
        return f'TextRecord("{shorten(self.text, MAX_REPR_LEN)}")'


class EventType:
    generic: str = 'generic'
    error: str = 'error'
    started: str = 'started'
    finished: str = 'finished'


class Event(Record):
    """
    Record produced by an internal event (usually error) inside the plugin
    """

    event_type: str = EventType.generic
    """text describing the nature of event, can be used to filter classes of events, such as errors"""
    text: str
    """text describing specific even details"""
    record: SerializeAsAny[Optional[Record]] = Field(exclude=True, default=None)
    """record that was being processed when this event happened"""

    def __str__(self):
        return self.text

    def __repr__(self):
        text = shorten(self.text, MAX_REPR_LEN)
        return f'Event(event_type="{self.event_type}", text="{text}")'

    def model_post_init(self, __context):
        if self.record is not None:
            self.origin = self.record.origin
            self.chain = self.record.chain


Subscription = Callable[[str, Record], None]
SubscriptionsMapping = Dict[str, List[Subscription]]


class MessageBus:
    PREFIX_IN = 'inputs'
    PREFIX_OUT = 'output'
    SEPARATOR = '/'

    HISTORY_SIZE = 20

    def __init__(self) -> None:
        self.subscriptions: SubscriptionsMapping = defaultdict(list)
        self.logger = logging.getLogger('bus')
        self.history: Dict[str, Deque[Record]] = defaultdict(lambda: deque(maxlen=self.HISTORY_SIZE))

    def sub(self, topic: str, callback: Subscription):
        self.logger.debug(f'subscription on topic {topic} by {callback!r}')
        self.subscriptions[topic].append(callback)

    def _generic_topic(self, specific_topic: str) -> str:
        direction, actor, entity, chain = self.split_subscription_topic(specific_topic)
        generic_topic = self.make_topic(direction, actor, entity, '')
        return generic_topic

    def pub(self, topic: str, message: Record):
        self.logger.debug(f'on topic {topic} message "{message!r}"')
        matching_callbacks = self.get_matching_callbacks(topic)
        for specific_topic, callbacks in matching_callbacks.items():
            if message.chain:
                targeted_message = message
            else:
                _, _, chain = self.split_message_topic(specific_topic)
                targeted_message = message.model_copy(deep=True)
                targeted_message.chain = chain
            for callback in callbacks:
                callback(specific_topic, targeted_message)

            generic_topic = self._generic_topic(specific_topic)
            self.add_to_history(generic_topic, targeted_message)
        if not matching_callbacks:
            # topic has no subscribers, meaning entity is not referenced in chains
            self.add_to_history(self._generic_topic(topic), message)

    def add_to_history(self, topic: str, message: Record):
        self.history[topic].append(message)

    def get_history(self, actor: str, entity: str, chain: str = '', direction: Literal['in', 'out'] = 'in') -> List[Record]:
        if direction == 'in':
            topic = self.incoming_topic_for(actor, entity, '')
        elif direction == 'out':
            topic = self.outgoing_topic_for(actor, entity, '')
        else:
            assert False, f'unexpected direction "{direction}"'
        if chain:
            records = [record for record in self.history[topic] if record.chain == chain]
        else:
            records = list(self.history[topic])
        records.sort(key = lambda record: record.created_at)
        return records

    def get_matching_callbacks(self, topic_pattern: str) -> SubscriptionsMapping:
        pattern_direction, pattern_actor, pattern_entity, pattern_chain = self.split_subscription_topic(topic_pattern)
        callbacks = defaultdict(list)
        for topic, callback in self.subscriptions.items():
            direction, actor, entity, chain = self.split_subscription_topic(topic)
            if pattern_direction != direction:
                continue
            if actor != pattern_actor:
                continue
            if entity != pattern_entity:
                continue
            if chain == '' or pattern_chain == '' or chain == pattern_chain:
                callbacks[topic].extend(callback)
        return callbacks

    def make_topic(self, *args: str):
        return self.SEPARATOR.join(args)

    def split_topic(self, topic: str):
        return topic.split(self.SEPARATOR)

    def incoming_topic_for(self, actor: str, entity: str, chain: str = '') -> str:
        return self.make_topic(self.PREFIX_IN, actor, entity, chain)

    def outgoing_topic_for(self, actor: str, entity: str, chain: str = '') -> str:
        return self.make_topic(self.PREFIX_OUT, actor, entity, chain)

    def split_message_topic(self, topic) -> Tuple[str, str, str]:
        _, actor, entity, chain = self.split_subscription_topic(topic)
        return actor, entity, chain

    def split_subscription_topic(self, topic) -> Tuple[str, str, str, str]:
        try:
            direction, actor, entity, chain = self.split_topic(topic)
        except ValueError:
            self.logger.error(f'failed to split message topic "{topic}"')
            raise
        return direction, actor, entity, chain

    def clear_subscriptions(self):
        self.subscriptions.clear()


class TerminatedAction(int, Enum):
    EXIT = 0
    RESTART = 2


class TasksController:
    class TerminatedError(KeyboardInterrupt):
        """Raised when application restart is requested"""

    _tasks: set[asyncio.Task] = set()

    def __init__(self, poll_interval: float = 5, logger: Optional[logging.Logger] = None) -> None:
        self.logger = logger or logging.getLogger('task_controller')
        self.poll_interval = poll_interval
        self.termination_pending = False
        self.termination_required = False
        self.terminated_action = TerminatedAction.EXIT
        self.tasks = self._tasks

    def create_task(self, coro: Coroutine, *, name: Optional[str] = None) -> asyncio.Task:
        task = asyncio.create_task(coro, name=name)
        if task in self.tasks:
            raise RuntimeError(f'newly created task {task} is already monitored')
        self.tasks.add(task)
        return task

    async def check_done_tasks(self, done: set[asyncio.Task]) -> None:
        for task in done:
            if not task.done():
                continue
            try:
                task_exception = task.exception()
                if task_exception is not None:
                    self.logger.error(f'task "{task.get_name()}" has terminated with exception', exc_info=task_exception)
            except asyncio.CancelledError:
                self.logger.debug(f'task "{task.get_name()}" cancelled')
            self.tasks.discard(task)

    async def monitor_tasks(self) -> None:
        while not self.termination_required:
            if not self.tasks:
                await asyncio.sleep(self.poll_interval)
                continue
            done, pending = await asyncio.wait(self.tasks, return_when=asyncio.FIRST_EXCEPTION, timeout=self.poll_interval)
            await self.check_done_tasks(done)
        self.termination_required = False
        raise self.TerminatedError()

    async def cancel_all_tasks(self) -> None:
        self.logger.debug(f'terminating {len(self.tasks)} tasks')
        while True:
            for task in self.tasks:
                task.cancel('terminating')
            done, pending = await asyncio.wait(self.tasks, return_when=asyncio.ALL_COMPLETED)
            await self.check_done_tasks(done)
            if not pending:
                break
            self.logger.debug(f'{len(pending)} more tasks left to terminate')
        self.logger.debug('all tasks terminated')

    async def terminate(self, delay: float, action: TerminatedAction):
        if self.termination_pending:
            self.logger.warning(f'active termination request is already pending')
            return
        self.termination_pending = True
        if delay > 0:
            self.logger.debug(f'terminating after {delay:.02f}')
            await asyncio.sleep(delay)
        self.logger.debug(f'terminating now')
        self.termination_pending = False
        self.terminated_action = action
        self.termination_required = True

    def terminate_after(self, delay: float, action: TerminatedAction):
        self.create_task(self.terminate(delay, action), name=f'terminate after {delay}')

    async def run_until_termination(self) -> TerminatedAction:
        try:
            await self.monitor_tasks()
        except self.TerminatedError:
            await self.cancel_all_tasks()
        except (asyncio.CancelledError, KeyboardInterrupt):
            await self.cancel_all_tasks()
            raise
        return self.terminated_action


class RuntimeContext:
    def __init__(self, bus: MessageBus, controller: TasksController):
        self.bus: MessageBus = bus
        self.controller: TasksController = controller
        self._sigint_handler = signal.getsignal(signal.SIGINT)
        self._sigterm_handler = signal.getsignal(signal.SIGINT)

    def _get_handler(self) -> Callable:
        controller = self.controller
        def handler(sig, frame):
            controller.terminate_after(0, TerminatedAction.EXIT)
        return handler

    def __enter__(self):
        signal.signal(signal.SIGINT, self._get_handler())
        signal.signal(signal.SIGTERM, self._get_handler())

    def __exit__(self, exc_type, exc_val, exc_tb):
        signal.signal(signal.SIGINT, self._sigint_handler)
        signal.signal(signal.SIGTERM, self._sigterm_handler)


    @classmethod
    def create(cls) -> 'RuntimeContext':
        bus = MessageBus()
        controller = TasksController()
        return cls(bus=bus, controller=controller)


class ActorConfig(BaseModel):
    model_config = ConfigDict(use_attribute_docstrings=True)

    name: str
    defaults: Dict[str, Any] = Field(default={}, exclude=True)


class ActorEntity(BaseModel):
    model_config = ConfigDict(use_attribute_docstrings=True)

    name: str
    """name of a specific entity. Used to reference it in `chains` section. Must be unique within a plugin"""
    reset_origin: bool = False
    """treat throughput records as if they have originated from this entity, 
    emitting them in every Chain this entity is used"""


class Actor(ABC):
    model_config = ConfigDict(use_attribute_docstrings=True)

    def __init__(self, conf: ActorConfig, entities: Sequence[ActorEntity], ctx: RuntimeContext):
        self.conf = conf
        self.logger = logging.getLogger(f'actor').getChild(conf.name)
        self.ctx = ctx
        self.bus = ctx.bus
        self.controller = ctx.controller
        self.entities: Dict[str, ActorEntity] = {entity.name: entity for entity in entities}

        for entity_name in self.entities:
            topic = self.bus.incoming_topic_for(self.conf.name, entity_name)
            self.bus.sub(topic, self._handle)

    def __repr__(self):
        text = f'{self.__class__.__name__}({list(self.entities)})'
        return shorten(text, MAX_REPR_LEN)

    def _handle(self, topic: str, record: Record) -> None:
        _, entity_name, _ = self.bus.split_message_topic(topic)
        if entity_name not in self.entities:
            logging.warning(f'received record on topic {topic}, but have no entity with name {entity_name} configured, dropping record {record!r}')
            return
        entity = self.entities[entity_name]
        try:
            self.handle_record(entity, record)
        except Exception:
            self.logger.exception(f'{self.conf.name}.{entity_name}: error while processing record "{record!r}"')

    @abstractmethod
    def handle_record(self, entity: ActorEntity, record: Record) -> None:
        '''Perform an action on record if entity in self.entities'''

    def on_record(self, entity: ActorEntity, record: Record):
        '''Implementation should call it for every new Record it produces'''
        if entity.reset_origin:
            record = record.model_copy(deep=True)
            record.chain = ''
        topic = self.bus.outgoing_topic_for(self.conf.name, entity.name, record.chain)
        self.bus.pub(topic, record)

    async def run(self):
        '''Will be run as asyncio task once everything is set up'''
        return


class MonitorEntity(ActorEntity):
    reset_origin: bool = Field(default=False, exclude=True)
    """excluded for Monitors because no good usecases"""


class Monitor(Actor, ABC):

    def __init__(self, conf: ActorConfig, entities: Sequence[MonitorEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)

    def handle_record(self, entity: MonitorEntity, record: Record) -> None:
        # For convenience’s sake monitors pass incoming records down the chain.
        # It allows using multiple monitors in a single chain, one after another
        self.on_record(entity, record)

    def on_record(self, entity: ActorEntity, record: Record):
        '''Implementation should call it for every new Record it produces'''
        origin = f'{self.conf.name}:{entity.name}'
        if record.origin == origin:
            self.logger.warning(f'[{entity.name}] received incoming record produced by self, which indicates loop in a chain, dropping: "{record!r}".\nCheck config for chains, passing records to each other in a loop.')
            return
        if record.origin is None:
            record.origin = origin
        super().on_record(entity, record)


class FilterEntity(ActorEntity):
    pass


class Filter(Actor):

    def __init__(self, conf: ActorConfig, entities: Sequence[FilterEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)
        self.logger = logging.getLogger(f'filters.{self.conf.name}')

    def handle_record(self, entity: FilterEntity, record: Record):
        filtered = self.match(entity, record)
        if filtered is not None:
            filtered.origin = filtered.origin or record.origin
            if filtered.chain is None:
                filtered.chain = ''
            elif not filtered.chain:
                filtered.chain = record.chain
            self.on_record(entity, filtered)
        else:
            self.logger.debug(f'[{entity.name}] record dropped: "{record!r}"')

    @abstractmethod
    def match(self, entity: FilterEntity, record: Record) -> Optional[Record]:
        '''
        Take a record and return it or a new/updated record
        if it matches some condition, otherwise return None.

        If returned record does not have "origin" field, it
        is copied from the original one. "chain" field is
        also copied if not set unless it is set to None, in
        which case it is set to empty string, making record
        propagate in all chains the filter is registered in.
        '''


class ActionEntity(ActorEntity):
    consume_record: bool = True
    """whether record should be consumed or passed down the chain after processing. Disabling it allows chaining multiple Actions"""
    timezone: Optional[str] = None
    """takes timezone name from <https://en.wikipedia.org/wiki/List_of_tz_database_time_zones> (or local time if omitted), converts record fields containing date and time to this timezone"""

    @field_validator('timezone')
    @classmethod
    def check_timezone(cls, timezone: Optional[str]) -> Optional[datetime.tzinfo]:
        return Timezone.get_tz(timezone)

    @field_serializer('timezone')
    @classmethod
    def serialize_timezone(cls, timezone: Optional[datetime.tzinfo]) -> Optional[str]:
        return Timezone.get_name(timezone)


class Action(Actor, ABC):

    def __init__(self, conf: ActorConfig, entities: Sequence[ActionEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)

    def handle_record(self, entity: ActionEntity, record: Record) -> None:
        record = record.as_timezone(entity.timezone)
        self.handle(entity, record)
        if not entity.consume_record:
            self.on_record(entity, record)

    @abstractmethod
    def handle(self, entity: ActionEntity, record: Record):
        '''Method for implementation to process incoming record'''


class Timezone:
    known: Dict[str, Any] = {}

    @classmethod
    def get_tz(cls, name: Optional[str]) -> Optional[datetime.tzinfo]:
        if name is None:
            return None
        tz = dateutil.tz.gettz(name)
        if tz is None:
            raise ValueError(f'Unknown timezone: {name}')
        cls.known[name] = tz
        return tz

    @classmethod
    def get_name(cls, tz: Optional[datetime.tzinfo]) -> Optional[str]:
        if tz is None:
            return None
        for name, timezone in cls.known.items():
            if tz == timezone:
                return name
        return tz.tzname(datetime.datetime.now())



import asyncio
import datetime
from asyncio.subprocess import Process
from typing import List, Optional

import pytest

from avtdl.core.interfaces import Record, RuntimeContext, TextRecord
from avtdl.plugins.execute.run_command import Command, CommandConfig, CommandEntity
from avtdl.plugins.rss.generic_rss import GenericRSSRecord


@pytest.fixture()
def text_record():
    return TextRecord(text='record1')


async def run_command(entity: CommandEntity, record: Record) -> Process:
    ctx = RuntimeContext.create()
    config = CommandConfig(name='test')
    actor = Command(config, [entity], ctx)

    flag = asyncio.Event()
    completed_process: Optional[Process] = None

    def done(process: Process, entity: CommandEntity, record: Record) -> None:
        nonlocal completed_process
        completed_process = process
        flag.set()

    actor.add_done_callback(done)

    actor.handle(entity, record)
    try:
        await asyncio.wait_for(flag.wait(), timeout=5)
    except asyncio.TimeoutError:
        pytest.fail('subprocess takes too long to complete')

    assert completed_process is not None
    return completed_process


@pytest.mark.asyncio
async def test_echo(capfd):
    text_record = TextRecord(text='record1')
    entity = CommandEntity(name='test', command='echo {text}')

    await run_command(entity, text_record)

    assert capfd.readouterr().out == 'record1\n'


@pytest.mark.asyncio
async def test_exit0(text_record):
    entity = CommandEntity(name='test', command='true')
    process = await run_command(entity, text_record)
    assert process.returncode == 0


@pytest.mark.asyncio
async def test_exit1(text_record):
    entity = CommandEntity(name='test', command='false')
    process = await run_command(entity, text_record)
    assert process.returncode == 1


class TestCommandArgs:

    @staticmethod
    def args_for(entity: CommandEntity, record: Record) -> List[str]:
        ctx = RuntimeContext.create()
        config = CommandConfig(name='test')
        actor = Command(config, [entity], ctx)
        args = actor.args_for(entity, record)
        return args

    @staticmethod
    @pytest.fixture()
    def text_record():
        return TextRecord(text='test test')

    @staticmethod
    @pytest.fixture()
    def feed_record():
        record = GenericRSSRecord(
            uid='1',
            url='https://example.com/1.html',
            author='example.com',
            title='#1',
            summary='about #1',
            published=datetime.datetime.now()
        )
        return record


class TestPlaceholders(TestCommandArgs):

    def test_text_placeholder(self, text_record):
        entity = CommandEntity(name='test', command='echo {text}')
        result = self.args_for(entity, text_record)
        assert result == ['echo', 'test test']

    def test_url_placeholder(self, feed_record):
        entity = CommandEntity(name='test', command='yt-dlp --cookies cookies.txt -f 220k/best {url}')
        result = self.args_for(entity, feed_record)
        assert result == ['yt-dlp', '--cookies', 'cookies.txt', '-f', '220k/best', 'https://example.com/1.html']

    def test_other_placeholders(self, feed_record):
        entity = CommandEntity(
            name='test',
            command="yt-dlp --add-header Referer:'https://example.com' {url} --output '[{author}] {title} ({uid}).%(ext)s'"
        )
        result = self.args_for(entity, feed_record)
        assert result == ['yt-dlp', '--add-header', 'Referer:https://example.com',
                          'https://example.com/1.html', '--output', '[example.com] #1 (1).%(ext)s']

    def test_missing_placeholder_unchanged(self, text_record):
        entity = CommandEntity(name='test', command='echo {text} {image}')
        result = self.args_for(entity, text_record)
        assert result == ['echo', 'test test', '{image}']

    def test_static_placeholders(self, text_record):
        entity = CommandEntity(name='test', command='echo {text} {image}',
                               static_placeholders={'{image}': '-o {text}.txt'})
        result = self.args_for(entity, text_record)
        assert result == ['echo', 'test test', '-o', 'test test.txt']


class TestFormatter(TestCommandArgs):

    def test_nested_quotes(self, text_record):
        entity = CommandEntity(name='test', command='''py -c "print('[hi there, {text}]'); exit(1)"''')
        result = self.args_for(entity, text_record)
        assert result == ['py', '-c', "print('[hi there, test test]'); exit(1)"]

    def test_escape_quotes(self, text_record):
        entity = CommandEntity(
            name='test',
            command='powershell "(New-Object -ComObject Wscript.Shell).Popup(\\"{text}\\", 0, \\"Done\\", 0x0)"'
        )
        result = self.args_for(entity, text_record)
        assert result == ['powershell', '(New-Object -ComObject Wscript.Shell).Popup("test test", 0, "Done", 0x0)']

    def test_datetime(self, feed_record):
        entity = CommandEntity(name='test', command="yt-dlp {url} --output '[%Y-%m-%d] {title}.%(ext)s'")
        now = datetime.datetime.now()
        result = self.args_for(entity, feed_record)
        assert result == ['yt-dlp', 'https://example.com/1.html',
                          '--output', f'[{now.year}-{now.month:02}-{now.day:02}] #1.%(ext)s']


import datetime
import json
import re
import urllib.request
from collections import defaultdict
from json import JSONDecodeError
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field


class VideoInfoError(ValueError):
    """Error fetching or parsing video info"""


class LoginRequiredError(VideoInfoError):
    """Raised when video page returns no data aside
    from {'playability_status': 'LOGIN_REQUIRED'},
    which usually indicated video being private"""


class VideoPlayabilityError(VideoInfoError):
    """Raised when video page returns {'playability_status': 'ERROR'}"""


class VideoFormat(BaseModel):
    itag: int
    url: str = Field(repr=False)
    mime: str
    quality: str


class VideoInfo(BaseModel):
    url: str
    title: str
    published: str
    uploaded: str
    author: str
    channel_id: str
    video_id: str
    summary: str = Field(repr=False)
    views: int
    length: int

    scheduled: Optional[datetime.datetime] = None
    live_start: Optional[datetime.datetime] = None
    live_end: Optional[datetime.datetime] = None

    is_unlisted: bool
    is_adult: bool
    is_livestream: bool
    is_upcoming: bool

    playability_status: str
    playability_reason: Optional[str] = None

    formats: Optional[List[VideoFormat]] = []


def get_video_page(url: str) -> str:
    with urllib.request.urlopen(url, timeout=15) as resp:
        data = resp.read()
        return data.decode('utf8')


def get_initial_player_response(page: str) -> dict:
    try:
        return get_initial_response_fast(page)
    except (ValueError, JSONDecodeError):
        return get_initial_response_slow(page)


def get_initial_response_fast(page: str) -> dict:
    re_initial_data = 'var ytInitialPlayerResponse = ([^;]*);'
    match = re.search(re_initial_data, page)
    if match is None:
        raise ValueError(f'Failed to find ytInitialPlayerResponse on the page')
    raw_data = match.groups()[0]
    data = json.loads(raw_data)
    return data


def get_initial_response_slow(page: str) -> dict:
    anchor = 'var ytInitialPlayerResponse = {'
    pos_start = page.find(anchor)
    if pos_start == -1:
        raise ValueError(f'Failed to find ytInitialPlayerResponse on the page')
    pos_start += len(anchor) - 1
    position = pos_start

    re_parenthesses = re.compile('[{}]')
    parenthesses_values = defaultdict(int, {'{': 1, '}': -1})
    parentheses = 0
    while True:
        parentheses += parenthesses_values[page[position]]
        if parentheses == 0:
            raw_data = page[pos_start:position + 1]
            response = json.loads(raw_data)
            return response
        position_match = re_parenthesses.search(page, position + 1)
        try:
            position = position_match.start()
        except AttributeError:
            raise ValueError(f'Failed to find matching set of parentheses after ytInitialPlayerResponse')


def get_embedded_player_response(page: str) -> dict:
    pos_start = page.find('{"embedded_player_response"')
    if pos_start == -1:
        raise ValueError(f'Failed to find embedded_player_response on page')
    parentheses = 1
    for position in range(pos_start + 1, len(page)):
        if page[position] == '{':
            parentheses += 1
        if page[position] == '}':
            parentheses -= 1
        if parentheses == 0:
            raw_data = page[pos_start:position + 1]
            response_data = json.loads(raw_data).get('embedded_player_response', '')
            response = json.loads(response_data)
            return response
    else:
        raise ValueError(f'Failed to find closing parenthesis for embedded_player_response')


def rename_keys(input_dict: Dict[str, Any], key_mapping: Dict[str, str]) -> Dict[str, Any]:
    output_dict = {}
    for data_key, input_key in key_mapping.items():
        value = input_dict.get(input_key)
        if value is not None:
            output_dict[data_key] = value
    return output_dict


def parse_playability_status(player_response: dict) -> Dict[str, Any]:
    playability_status = player_response.get('playabilityStatus')
    if playability_status is None:
        return {}

    info = rename_keys(playability_status, {'playability_status': 'status', 'playability_reason': 'reason'})

    try:
        date = playability_status["liveStreamability"]["liveStreamabilityRenderer"]["offlineSlate"][
            "liveStreamOfflineSlateRenderer"]["scheduledStartTime"]
        date = datetime.datetime.fromtimestamp(int(date), tz=datetime.timezone.utc)
    except (KeyError, TypeError, ValueError):
        pass
    else:
        info['scheduled'] = date

    return info


def parse_video_details(player_response: dict) -> Dict[str, Any]:
    video_details = player_response.get('videoDetails')
    if video_details is None:
        return {}
    key_mapping = {
        'title': 'title',
        'author': 'author',
        'channel_id': 'channelId',
        'video_id': 'videoId',
        'summary': 'shortDescription',
        'views': 'viewCount',
        'length': 'lengthSeconds',
        'is_livestream': 'isLiveContent',
    }
    info = rename_keys(video_details, key_mapping)
    info['is_upcoming'] = video_details.get('isUpcoming', False)
    return info


def parse_microformat(player_response: dict) -> Dict[str, Any]:
    microformat = player_response.get('microformat', {}).get('playerMicroformatRenderer')
    if microformat is None:
        return {}
    key_mapping = {
        'published': 'publishDate',
        'uploaded': 'uploadDate',
        'author': 'ownerChannelName',
        'views': 'viewCount',
        'length': 'lengthSeconds',
        'is_unlisted': 'isUnlisted',
    }
    info = rename_keys(microformat, key_mapping)
    if microformat.get('title') is not None:
        info['title'] = microformat.get('title', {}).get('simpleText')
    if microformat.get('description') is not None:
        info['summary'] = microformat.get('description', {}).get('simpleText')
    info['is_adult'] = not microformat.get('isFamilySafe')
    info['is_livestream'] = microformat.get('liveBroadcastDetails') is not None

    live_details = microformat.get('liveBroadcastDetails')
    if live_details is not None:
        if live_details.get('endTimestamp') is not None:  # live ended
            info['live_start'] = live_details.get('startTimestamp')
            info['live_end'] = live_details.get('endTimestamp')
        elif live_details.get('isLiveNow') == True:  # live is live
            info['live_start'] = live_details.get('startTimestamp')
        else:  # live is scheduled
            date = live_details.get('startTimestamp')
            try:
                date = datetime.datetime.fromisoformat(date)
                info['scheduled'] = date
            except (ValueError, TypeError):
                pass

    return info


def parse_video_formats(player_response: dict) -> List[VideoFormat]:
    formats = player_response.get('streamingData')
    if formats is None:
        return []
    formats_list = []
    formats_list.extend(formats.get('formats', []))
    formats_list.extend(formats.get('adaptiveFormats', []))

    parsed_formats = []
    for item in formats_list:
        try:
            itag = item['itag']
            url = item['url']
            mime = item['mimeType']
            quality = item.get('qualityLabel') or item['quality']
            parsed_formats.append(VideoFormat(itag=itag, url=url, mime=mime, quality=quality))
        except (TypeError, KeyError):
            continue
    return parsed_formats


def parse_player_response(player_response: dict) -> Dict[str, Any]:
    info = {}
    info.update(parse_video_details(player_response))
    info.update(parse_microformat(player_response))
    info.update(parse_playability_status(player_response))
    info['formats'] = parse_video_formats(player_response)
    return info


def parse_video_page(page: str, url: str) -> VideoInfo:
    response = get_initial_player_response(page)
    data = parse_player_response(response)
    data['url'] = url
    if data['playability_status'] == 'LOGIN_REQUIRED':
        reason = data.get('playability_reason') or f'Video might be private. Raw data: {data}'
        raise LoginRequiredError(f'Video info for {url} unavailable: login required. {reason}')
    if data['playability_status'] == 'ERROR':
        raise VideoPlayabilityError(f'Video was deleted or never existed: {url} . Raw data: {data}')
    info = VideoInfo(**data)
    return info


def get_video_info(url: str) -> VideoInfo:
    page = get_video_page(url)
    return parse_video_page(page, url)



import asyncio
import datetime
import shlex
from asyncio.subprocess import Process
from hashlib import sha1
from pathlib import Path
from typing import Callable, Dict, List, Optional, Sequence

from pydantic import field_validator

from avtdl.core import utils
from avtdl.core.config import Plugins
from avtdl.core.interfaces import Action, ActionEntity, ActorConfig, Event, EventType, Record, RuntimeContext
from avtdl.core.utils import Fmt, check_dir, sanitize_filename

Plugins.register('execute', Plugins.kind.ASSOCIATED_RECORD)(Event)


@Plugins.register('execute', Plugins.kind.ACTOR_CONFIG)
class CommandConfig(ActorConfig):
    pass

@Plugins.register('execute', Plugins.kind.ACTOR_ENTITY)
class CommandEntity(ActionEntity):
    command: str
    """shell command to be executed on every received record. Supports placeholders that will be replaced with currently processed record fields values"""
    working_dir: Optional[Path] = None
    """path to the directory where command will be executed. If not set current working directory is used. Supports templating with {...}"""
    log_dir: Optional[Path] = None
    """write executed process output to a file in this directory if set. If it is not set, output will not be redirected to file"""
    log_filename: Optional[str] = None
    """filename to write executed process output to. If not defined, it is generated automatically based on command and entity name"""
    static_placeholders: Dict[str, str] = {}
    """parts of `command` string that will be replaced with provided values, defined as mapping `'placeholder': 'replacement string'`. Intended to allow reusing same `command` template for multiple entities"""
    forward_failed: bool = False
    """emit currently processed record down the chain if the subprocess returned non-zero exit code. Can be used to define fallback command in case this one fails"""
    report_failed: bool = True
    """emit Event with type "error" if the subprocess returned non-zero exit code or raised exception"""
    report_finished: bool = False
    """emit Event with type "finished" if the subprocess returned zero as exit code"""
    report_started: bool = False
    """emit Event with type "started" before starting a subprocess"""

    @field_validator('log_dir')
    @classmethod
    def check_dir(cls, path: Optional[Path]):
        if path is None:
            return path
        ok = utils.check_dir(path)
        if ok:
            return path
        else:
            raise ValueError(f'check path "{path}" exists and is a writeable directory')

    @field_validator('command')
    @classmethod
    def split_args(cls, command: str):
        _ = shlex.split(command) # might raise ValueError
        return command


@Plugins.register('execute', Plugins.kind.ACTOR)
class Command(Action):
    """
    Run pre-defined shell command

    Take `command` string, replace keywords wrapped in curly braces with corresponding fields
    of currently processed record. For example, if `command` is set to

        "yt-dlp {url}"`

    and currently processed record comes from Youtube RSS feed and has `url` field value
    `https://www.youtube.com/watch?v=L692Sxz3thw`, then with default `placeholders`
    resulting command will be

        yt-dlp https://www.youtube.com/watch?v=L692Sxz3thw

    See `Formatting templates` section of README.md for more detailed description
    of the formatting options.

    `command` string is not treated as raw shell command. Instead, it is split into list
     of elements, where first element specifies the program executable, and the rest
     specify the arguments. It is therefore not possible to use shell features such as pipes
     or execute multiple commands in one line.

    Make sure the executable the command uses (`yt-dlp` in this case) is installed and
    can be run from the working directory by current user. It is advised to confirm that
    the command can be executed manually and it finishes without errors before automating it.

    For each entity, a separate working directory can be configured. Output is shown
    in the same window by default, but can be redirected to a file, with either static
    or autogenerated name.

    Produces Events at startup and successful or erroneous termination of the executed command
    if corresponding entity settings are enabled. They can be used to send Discord
    or Jabber notifications or execute another command when it happens.

    Processed record itself can also be passed down the chain if the command failed,
    providing a way to try a different one as a fallback. For example, record with
    Youtube url could be first handled by `ytarchive` and passed to `yt-dlp` if
    it happens to fail due to video link not being a livestream.
    """

    def __init__(self, conf: CommandConfig, entities: Sequence[CommandEntity], ctx: RuntimeContext):
        super().__init__(conf, entities, ctx)
        self.running_commands: Dict[str, asyncio.Task] = {}
        self.done_callbacks: List[Callable[[Process, CommandEntity, Record], None]] = []

    def handle(self, entity: CommandEntity, record: Record):
        self.add(entity, record)

    def args_for(self, entity: CommandEntity, record: Record):
        '''
        Fill entity.command with placeholders

        - fill static_placeholders
        - split into args
        - fill each argument using generic Fmt.format
        '''
        command = entity.command
        for placeholder, static_value in entity.static_placeholders.items():
            command = command.replace(placeholder, static_value)
        try:
            args = shlex.split(command)
        except ValueError as e:
            self.logger.error(f'{self.conf.name}: error parsing "command" field of entity "{entity.name}" with value "{entity.command}": {e}')
            raise
        new_args = []
        for arg in args:
            new_arg = Fmt.format(arg, record, tz=entity.timezone)
            new_args.append(new_arg)
        return new_args

    @staticmethod
    def shell_for(args: List[str]) -> str:
        return shlex.join(args)

    def _generate_task_id(self, entity: CommandEntity, record: Record, command_line: str) -> str:
        record_hash = record.hash()
        task_id = f'Task for {entity.name}: on record {record!r} ({record_hash}) executing "{command_line}"'
        return task_id

    def add(self, entity: CommandEntity, record: Record):
        args = self.args_for(entity, record)
        if entity.working_dir is None:
            working_dir = Path.cwd()
            self.logger.info(f'[{entity.name}] working directory is not specified, using current directory instead: {working_dir}')
        else:
            working_dir = Fmt.format_path(entity.working_dir, record, tz=entity.timezone)
            ok = check_dir(working_dir)
            if not ok:
                self.logger.warning(f'[{entity.name}] check if working directory "{working_dir}" exists and is a writeable directory')

        command_line = self.shell_for(args)
        task_id = self._generate_task_id(entity, record, command_line)
        if task_id in self.running_commands:
            msg = f'[{entity.name}] command "{command_line}" for record {record!r} is already running, will not call again'
            self.logger.info(msg)
            return
        self.logger.debug(f'[{entity.name}] executing command "{command_line}" for record {record!r}')
        task = self.run_subprocess(args, task_id, working_dir, entity, record)
        self.running_commands[task_id] = asyncio.get_event_loop().create_task(task, name=f'{self.conf.name}:{entity.name}:{task_id}')

        self._check_running_commands()

    def _check_running_commands(self):
        for name, task in self.running_commands.items():
            if task.done() and task.exception() is not None:
                self.logger.error(f'[{name}] task {task.get_name()} has terminated with exception', exc_info=task.exception())

    def _get_output_file(self, entity: CommandEntity, record: Record, task_id: str) -> Optional[Path]:
        if entity.log_dir is None:
            return None
        ok = check_dir(entity.log_dir)
        if not ok:
            self.logger.warning(f'[{entity.name}] check if directory specified in "output_dir" value "{entity.log_dir}" exists and is a writeable directory')
            self.logger.warning(f'[{entity.name}] output of running command will be redirected to stdout')
            return None
        if entity.log_filename is not None:
            filename = Fmt.format(entity.log_filename, record, tz=entity.timezone)
        else:
            timestamp = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S.%f')
            command_pre_hash = sha1(task_id.encode())
            command_hash = command_pre_hash.hexdigest()
            filename = f'command_{entity.name}_{timestamp}_{command_hash[:6]}_stdout.log'
        filename = sanitize_filename(filename)
        return entity.log_dir / filename

    async def run_subprocess(self, args: List[str], task_id: str, working_dir: Path, entity: CommandEntity, record: Record):
        command_line = self.shell_for(args)
        self.logger.info(f'[{entity.name}] executing command {command_line}')
        if entity.report_started:
            event = Event(event_type=EventType.started, text=f'Running command: {command_line}', record=record)
            self.on_record(entity, event)
        stdout_path = self._get_output_file(entity, record, task_id)
        try:
            stdout = open(stdout_path, 'at', encoding='utf8') if stdout_path is not None else None
        except OSError as e:
            self.logger.warning(f'[{entity.name}] failed to open file {stdout_path}: {e}. Command output will not be written')
            stdout = None
        try:
            if stdout is None:
                process = await asyncio.create_subprocess_exec(*args, cwd=working_dir)
            else:
                with stdout:
                    stdout.write(f'# [{self.conf.name}.{entity.name}] > {command_line}\n')
                    stdout.flush()
                    process = await asyncio.create_subprocess_exec(*args, cwd=working_dir, stdout=stdout, stderr=stdout)
        except Exception as e:
            self.logger.exception(f'[{entity.name}] failed to execute command "{command_line}": {e}')
            if entity.report_failed:
                text = f'[{entity.name}] failed to execute command: {command_line}'
                event = Event(event_type=EventType.error, text=text, record=record)
                self.on_record(entity, event)
            if entity.forward_failed:
                self.on_record(entity, record)
            return
        try:
            await process.wait()
        except (KeyboardInterrupt, asyncio.CancelledError):
            self.logger.warning(f'[{entity.name}] application is terminating before running command has completed. Check on it and restart manually if needed. Process PID: {process.pid}. Exact command line:\n{command_line}')
            raise

        self.running_commands.pop(task_id)

        self.logger.debug(f'[{entity.name}] subprocess for {command_line} finished with exit code {process.returncode}')

        if process.returncode == 0:
            if entity.report_finished:
                text = f'[{entity.name}] command finished successfully: {command_line}'
                event = Event(event_type=EventType.finished, text=text, record=record)
                self.on_record(entity, event)
        else:
            if entity.report_failed:
                text = f'[{entity.name}] command finished with error: {command_line}'
                event = Event(event_type=EventType.error, text=text, record=record)
                self.on_record(entity, event)
            if entity.forward_failed:
                self.on_record(entity, record)

        for cb in self.done_callbacks:
            cb(process, entity, record)

    def add_done_callback(self, callback: Callable[[Process, CommandEntity, Record], None]):
        """register function to be called on every completed subprocess"""
        self.done_callbacks.append(callback)